# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 11:28
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: logging_config.py

from typing import Optional, List, Dict

import tomli


def _load_toml_file(path: str) -> dict:
    """Load the toml file.

    Args:
        path(str): the path of the toml file
    Returns:
        dict: the value of the toml file
    """
    with open(path, 'rb') as f:
        config_data = tomli.load(f)
    return config_data


class LoggingConfig(object):
    """Config class of logging utils."""
    log_format: str = ("<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> "
                       "| <level>{level: <8}</level> "
                       "| {extra[context_prefix]} "
                       "| <cyan>{name}</cyan>"
                       ":<cyan>{function}</cyan>"
                       ":<cyan>{line}</cyan> "
                       "| <level>{message}</level>")
    log_level: str = "INFO"
    log_extend_module_list: List[str] = ["sls_log"]
    log_extend_module_switch: Dict[str, bool] = {}
    log_path: Optional[str] = None
    log_rotation: str = "10 MB"
    log_retention: str = "3 days"

    # Aliyun sls configs.
    sls_endpoint: str = ""
    sls_project: str = ""
    sls_log_store: str = ""
    access_key_id: str = ""
    access_key_secret: str = ""
    sls_log_queue_max_size: int = 1000
    sls_log_send_interval: float = 3.0

    def __init__(self, config_path: Optional[str] = None):
        """Initialize the log config using config file if specified, otherwise
        use default value.

        Args:
            config_path(str):
                the path of the toml file
        """
        self.__config = None
        try:
            self.__config = _load_toml_file(config_path)["LOG_CONFIG"]
        except (FileNotFoundError, TypeError):
            print("can't find log config file, use default config")
            for log_module in LoggingConfig.log_extend_module_list:
                LoggingConfig.log_extend_module_switch[log_module] = False
            return
        except (tomli.TOMLDecodeError, KeyError):
            print("log config file isn't a valid toml, use default config.")
            for log_module in LoggingConfig.log_extend_module_list:
                LoggingConfig.log_extend_module_switch[log_module] = False
            return

        for log_module in LoggingConfig.log_extend_module_list:
            switch = self._get_config_or_default("EXTEND_MODULE",
                                                 log_module)
            if switch and switch.lower() == 'true':
                LoggingConfig.log_extend_module_switch[log_module] = True
            else:
                LoggingConfig.log_extend_module_switch[log_module] = False

        log_level = self._get_config_or_default("BASIC_CONFIG",
                                                "log_level")
        if log_level:
            LoggingConfig.log_level = log_level.upper()

        log_path = self._get_config_or_default("BASIC_CONFIG",
                                               "log_path")
        if log_path:
            LoggingConfig.log_path = log_path

        log_rotation = self._get_config_or_default("BASIC_CONFIG",
                                                   "log_rotation")
        if log_rotation:
            LoggingConfig.log_rotation = log_rotation

        log_retention = self._get_config_or_default("BASIC_CONFIG",
                                                    "log_retention")
        if log_retention:
            LoggingConfig.log_retention = log_retention

        # Read sls config when sls extend module come into effect.
        if LoggingConfig.log_extend_module_switch["sls_log"]:
            LoggingConfig.sls_endpoint = self._get_config_or_default(
                "ALIYUN_SLS_CONFIG", "sls_endpoint")
            LoggingConfig.sls_project = self._get_config_or_default(
                "ALIYUN_SLS_CONFIG", "sls_project")
            LoggingConfig.sls_log_store = self._get_config_or_default(
                "ALIYUN_SLS_CONFIG", "sls_log_store")
            LoggingConfig.access_key_id = self._get_config_or_default(
                "ALIYUN_SLS_CONFIG", "access_key_id")
            LoggingConfig.access_key_secret = self._get_config_or_default(
                "ALIYUN_SLS_CONFIG", "access_key_secret")
            LoggingConfig.sls_log_queue_max_size = int(
                self._get_config_or_default("ALIYUN_SLS_CONFIG",
                                            "sls_log_queue_max_size"))
            LoggingConfig.sls_log_send_interval = float(
                self._get_config_or_default("ALIYUN_SLS_CONFIG",
                                            "sls_log_send_interval"))

    def _get_config_or_default(self, section, key, default_value=None):
        """Get config attribute from toml data, return default_value if no such
        attribute."""
        try:
            return self.__config[section][key]
        except KeyError:
            return default_value


def init_log_config(config_path: Optional[str] = None):
    LoggingConfig(config_path)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/13 10:42
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: planner.py
"""Base class for Planner."""
from abc import abstractmethod
import logging
from queue import Queue
from typing import Optional, List, Any

from langchain_core.runnables import RunnableSerializable

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.message import Message
from agentuniverse.agent.memory.memory_manager import MemoryManager
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.component_configer.configers.planner_configer import PlannerConfiger
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.base.util.memory_util import generate_messages, get_memory_string

logging.getLogger().setLevel(logging.ERROR)


class Planner(ComponentBase):
    """
    Base class for all planners.

    All planners should inherit from this class
    """
    name: Optional[str] = None
    description: Optional[str] = None
    output_key: str = 'output'
    input_key: str = 'input'
    prompt_assemble_order: list = ['introduction', 'target', 'instruction']

    def __init__(self):
        """Initialize the ComponentBase."""
        super().__init__(component_type=ComponentEnum.PLANNER)

    @abstractmethod
    def invoke(self, agent_model: AgentModel, planner_input: dict,
               input_object: InputObject) -> dict:
        """Invoke the planner.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): The input parameters passed by the user.
        Returns:
            dict: The planner result.
        """
        pass

    def handle_memory(self, agent_model: AgentModel, planner_input: dict) -> Memory | None:
        """Memory module processing.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
        Returns:
             Memory: The memory.
        """
        chat_history: list = planner_input.get('chat_history')
        memory_name = agent_model.memory.get('name')

        # get memory instance
        memory: Memory = MemoryManager().get_instance_obj(component_instance_name=memory_name)
        if memory is None:
            return None

        # generate a list of temporary messages from the given chat history and add them to the memory instance.
        temporary_messages: list[Message] = generate_messages(chat_history)
        if temporary_messages:
            memory.add(temporary_messages, **planner_input)

        llm_name = agent_model.profile.get('llm_model', {}).get('name')
        llm: LLM = LLMManager().get_instance_obj(llm_name)

        params: dict = dict()
        params['llm'] = llm
        params['agent_llm_name'] = llm_name
        return memory.set_by_agent_model(**params)

    def run_all_actions(self, agent_model: AgentModel, planner_input: dict, input_object: InputObject):
        """Tool and knowledge processing.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): Agent input object.
        """
        action: dict = agent_model.action or dict()
        tools: list = action.get('tool') or list()
        knowledge: list = action.get('knowledge') or list()
        agents: list = action.get('agent') or list()

        action_result: list = list()

        for tool_name in tools:
            tool = ToolManager().get_instance_obj(tool_name)
            if tool is None:
                continue
            tool_input = {key: input_object.get_data(key) for key in tool.input_keys}
            action_result.append(tool.run(**tool_input))

        for knowledge_name in knowledge:
            knowledge: Knowledge = KnowledgeManager().get_instance_obj(knowledge_name)
            if knowledge is None:
                continue
            knowledge_res: List[Document] = knowledge.query_knowledge(
                query_str=input_object.get_data(self.input_key),
                **input_object.to_dict()
            )
            action_result.append(knowledge.to_llm(knowledge_res))

        for agent_name in agents:
            agent = AgentManager().get_instance_obj(agent_name)
            if agent is None:
                continue
            agent_input = {key: input_object.get_data(key) for key in agent.input_keys()}
            output_object = agent.run(**agent_input)
            action_result.append("\n".join([output_object.get_data(key)
                                            for key in agent.output_keys()
                                            if output_object.get_data(key) is not None]))

        planner_input['background'] = planner_input['background'] or '' + "\n".join(action_result)

    def handle_prompt(self, agent_model: AgentModel, planner_input: dict):
        """Prompt module processing.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
        Returns:
            Prompt: The prompt instance.
        """
        pass

    def handle_llm(self, agent_model: AgentModel) -> LLM:
        """Language model module processing.

        Args:
            agent_model (AgentModel): Agent model object.
        Returns:
            LLM: The language model.
        """
        llm_name = agent_model.profile.get('llm_model').get('name')
        llm: LLM = LLMManager().get_instance_obj(component_instance_name=llm_name)
        return llm

    def initialize_by_component_configer(self, component_configer: PlannerConfiger) -> 'Planner':
        """Initialize the planner by the PlannerConfiger object.

        Args:
            component_configer(PlannerConfiger): the PlannerConfiger object
        Returns:
            Planner: the planner object
        """
        self.name = component_configer.name
        self.description = component_configer.description
        self.input_key = component_configer.input_key or self.input_key
        self.output_key = component_configer.output_key or self.output_key
        return self

    @staticmethod
    def stream_output(input_object: InputObject, data: dict):
        """Stream output.

        Args:
            input_object (InputObject): Agent input object.
            data (dict): The data to be streamed.
        """
        output_stream: Queue = input_object.get_data('output_stream', None)
        if output_stream is None:
            return
        output_stream.put_nowait(data)

    def invoke_chain(self, agent_model: AgentModel, chain: RunnableSerializable[Any, str], planner_input: dict,
                     chat_history,
                     input_object: InputObject):

        if not input_object.get_data('output_stream'):
            res = chain.invoke(input=planner_input, config={"configurable": {"session_id": "unused"}})
            return res
        result = []
        for token in chain.stream(input=planner_input, config={"configurable": {"session_id": "unused"}}):
            self.stream_output(input_object, {
                'type': 'token',
                'data': {
                    'chunk': token,
                    'agent_info': agent_model.info
                }
            })
            result.append(token)
        return "".join(result)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/25 14:35
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: nl2api_agent.py
from agentuniverse.agent.template.nl2api_agent_template import Nl2ApiAgentTemplate


class Nl2ApiAgent(Nl2ApiAgentTemplate):
    """Nl2Api Agent class."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 15:39
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: config_type_enum.py
from enum import Enum


class ConfigTypeEnum(Enum):
    """The enumeration of the supported configuration file types."""
    TOML = "toml"
    YAML = "yaml"
    JSON = "json"
    XML = "xml"
    PROPERTIES = "properties"
    INI = "ini"
    ENV = "env"



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/11 17:49
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: local_memory_storage.py
from typing import Optional, List, Dict

from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.agent.memory.message import Message


class LocalMemoryStorage(MemoryStorage):
    """The local temporary memory storage class.

    Attributes:
        messages (dict[str, dict[str, list[Message]]]): The messages in the local temporary memory.
    """

    messages: Optional[Dict[str, Dict[str, List[Message]]]] = dict()

    def add(self, message_list: List[Message], session_id: str = '', agent_id: str = '', **kwargs) -> None:
        """Add messages to the memory db.

        Args:
            message_list (List[Message]): The list of messages to add.
            session_id (str): The session id of the memory to add.
            agent_id (str): The agent id of the memory to add.
        """
        if not message_list:
            return
        session = self.messages.setdefault(session_id, {})
        agent_messages = session.setdefault(agent_id, [])
        agent_messages.extend(message_list)

    def delete(self, session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Delete the memory from the database.

        Args:
            session_id (str): The session id of the memory to delete.
            agent_id (str): The agent id of the memory to delete.
        """
        if session_id is not None:
            if agent_id is None:
                self.messages.pop(session_id, None)
            else:
                self.messages.get(session_id, {}).pop(agent_id, None)

    def get(self, session_id: str = '', agent_id: str = '', top_k=10, **kwargs) -> \
            List[Message]:
        """Get messages from the memory db.

        Args:
            session_id (str): The session id of the memory to get.
            agent_id (str): The agent id of the memory to get.
            top_k (int): The number of messages to get.

        Returns:
            List[Message]: The list of aU messages.
        """
        memories = self.messages.get(session_id, {}).get(agent_id, [])
        return memories[-top_k:]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/30 20:30
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: common_util.py
import os
from pathlib import Path


def dict_does_not_contain_keys(d: dict, keys: list) -> bool:
    """Check if the dictionary does not contain any of the specified keys."""
    return all(key not in d for key in keys)


def get_core_path():
    if os.path.exists(os.path.join('..', 'core')):
        return Path(os.path.join('..', 'core'))
    elif os.path.exists(os.path.join('..', '..', 'app', 'core')):
        return Path(os.path.join('..', '..', 'app', 'core'))
    return None


def get_resources_path():
    if os.path.exists(os.path.join('..', '..', 'platform', 'difizen', 'resources')):
        return Path(os.path.join('..', '..', 'platform', 'difizen', 'resources'))
    elif os.path.exists(os.path.join('..', 'resources')):
        return Path(os.path.join('..', 'resources'))
    else:
        return Path(os.path.join('..', '..', 'app', 'resources'))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 21:53
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: message_do.py
import datetime
from typing import Optional

from pydantic import BaseModel, Field


class MessageDO(BaseModel):
    id: Optional[int] = Field(description="ID", default=None)
    session_id: str = Field(description="Session id")
    ext_info: Optional[dict] = Field(description="Message ext info.", default={})
    content: Optional[str] = Field(description="Message content", default='')
    gmt_created: Optional[datetime.datetime] = Field(
        description="Create time", default_factory=datetime.datetime.now)
    gmt_modified: Optional[datetime.datetime] = Field(
        description="Modified time", default_factory=datetime.datetime.now)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 19:42
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: node.py
from abc import abstractmethod
from typing import Optional, Dict, List, Any

from pydantic import BaseModel

from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.node.node_config import NodeOutputParams, NodeInputParams
from agentuniverse.workflow.workflow_output import WorkflowOutput


class NodeData(BaseModel):
    outputs: Optional[List[NodeOutputParams]] = None


class Node(BaseModel):
    """The basic class of the node."""

    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    type: NodeEnum = None
    workflow_id: Optional[str] = None
    position: Optional[dict] = None
    _data: Optional[NodeData] = None
    _data_cls = NodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._data = self._data_cls(**kwargs.get('data', {}))

    @abstractmethod
    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        raise NotImplementedError

    def run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        return self._run(workflow_output)

    @staticmethod
    def _resolve_input_params(input_params: List[NodeInputParams],
                              workflow_output: WorkflowOutput) -> Dict[str, Any]:
        """Resolve the input parameters of the node.

        Args:
            input_params (List[NodeInputParams]): The input parameters of the node.
            workflow_output (WorkflowOutput): The output of the workflow.
        """
        node_input_params = {}
        for input_param in input_params:
            val = input_param.value
            if val.type == 'reference':
                reference_node_id = val.content[0]
                reference_output_params: List[NodeOutputParams] = workflow_output.workflow_parameters.get(
                    reference_node_id, [])
                node_input_params[input_param.name] = next(
                    (param.value for param in reference_output_params if param.name == val.content[1]), None)
            else:
                node_input_params[input_param.name] = val.content
        return node_input_params



from ..base.annotation.singleton import singleton
from ..base.component.component_enum import ComponentEnum
from ..base.component.component_manager_base import ComponentManagerBase
from .service import Service


@singleton
class ServiceManager(ComponentManagerBase[Service]):
    """A singleton manager class of the service."""

    def __init__(self):
        super().__init__(ComponentEnum.SERVICE)




# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/14 16:02
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: planning_planner.py
"""Planning planner module."""
from langchain_core.output_parsers import StrOutputParser

from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel


class PlanningPlanner(Planner):
    """Planning planner class."""

    def invoke(self, agent_model: AgentModel, planner_input: dict,
               input_object: InputObject) -> dict:
        """Invoke the planner.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): The input parameters passed by the user.
        Returns:
            dict: The planner result.
        """
        memory: Memory = self.handle_memory(agent_model, planner_input)

        llm: LLM = self.handle_llm(agent_model)

        prompt: Prompt = self.handle_prompt(agent_model, planner_input)
        process_llm_token(llm, prompt.as_langchain(), agent_model.profile, planner_input)

        memory_messages = assemble_memory_input(memory, planner_input)
        chain = prompt.as_langchain() | llm.as_langchain_runnable(agent_model.llm_params()) | StrOutputParser()

        res = self.invoke_chain(agent_model, chain, planner_input, None, input_object)

        memory_messages = assemble_memory_output(memory=memory,
                                                 agent_input=planner_input,
                                                 content=f"Human: {planner_input.get(self.input_key)}, AI: {res}",
                                                 memory_messages=memory_messages)
        return {**planner_input, self.output_key: res, 'chat_history': memory_messages}

    def handle_prompt(self, agent_model: AgentModel, planner_input: dict) -> Prompt:
        """Prompt module processing.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
        Returns:
            Prompt: The prompt instance.
        """
        expert_framework = planner_input.pop('expert_framework', '') or ''

        profile: dict = agent_model.profile

        profile_instruction = profile.get('instruction')
        profile_instruction = expert_framework + profile_instruction if profile_instruction else profile_instruction

        profile_prompt_model: AgentPromptModel = AgentPromptModel(introduction=profile.get('introduction'),
                                                                  target=profile.get('target'),
                                                                  instruction=profile_instruction)

        # get the prompt by the prompt version
        prompt_version: str = profile.get('prompt_version')
        version_prompt: Prompt = PromptManager().get_instance_obj(prompt_version)

        if version_prompt is None and not profile_prompt_model:
            raise Exception("Either the `prompt_version` or `introduction & target & instruction`"
                            " in agent profile configuration should be provided.")
        if version_prompt:
            version_prompt_model: AgentPromptModel = AgentPromptModel(
                introduction=getattr(version_prompt, 'introduction', ''),
                target=getattr(version_prompt, 'target', ''),
                instruction=expert_framework + getattr(version_prompt, 'instruction', ''))
            profile_prompt_model = profile_prompt_model + version_prompt_model

        return Prompt().build_prompt(profile_prompt_model, self.prompt_assemble_order)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/16 14:33
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: enum.py
import enum
from enum import Enum


@enum.unique
class PromptProcessEnum(Enum):
    TRUNCATE = 'truncate'
    STUFF = 'stuff'
    MAP_REDUCE = 'map_reduce'

    @classmethod
    def from_value(cls, value):
        for member in cls:
            if member.value.lower() == value.lower():
                return member
        raise ValueError(f"No enum member with value: {value}")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/24 10:59
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: query_paraphraser.py

from abc import abstractmethod
from typing import Optional

from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class QueryParaphraser(ComponentBase):
    """The basic class for query paraphraser."""

    component_type: ComponentEnum = ComponentEnum.QUERY_PARAPHRASER
    name: Optional[str] = None
    description: Optional[str] = None

    @abstractmethod
    def query_paraphrase(self, origin_query: Query) -> Query:
        """Paraphrase the origin query string to different style."""


    def _initialize_by_component_configer(self,
                                         query_paraphraser_config: ComponentConfiger) \
            -> 'QueryParaphraser':
        """Initialize the QueryParaphraser by the ComponentConfiger object.

        Args:
            query_paraphraser_config(ComponentConfiger): A configer contains query_paraphraser
            basic info.
        Returns:
            QueryParaphraser: A query_paraphraser instance.
        """
        if query_paraphraser_config.name:
            self.name = query_paraphraser_config.name
        if query_paraphraser_config.description:
            self.description = query_paraphraser_config.description

        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/25 17:32
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_util.py
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.util.memory_util import get_memory_string


def assemble_memory_input(memory: Memory, agent_input: dict) -> list[Message]:
    """Assemble memory information for the agent input parameters.

    Args:
        memory (Memory): The memory instance.
        agent_input (dict): Agent input parameters for the agent.

    Returns:
        list[Message]: The retrieved memory messages.
    """
    memory_messages = []
    if memory:
        # get the memory messages from the memory instance.
        memory_messages = memory.get(**agent_input)
        # convert the memory messages to a string and add it to the agent input object.
        memory_str = get_memory_string(memory_messages)
        agent_input[memory.memory_key] = memory_str
    return memory_messages


def assemble_memory_output(memory: Memory, agent_input: dict,
                           content: str, source: str = None, memory_messages=None) -> \
        list[Message]:
    """Assemble the historical memory information and current memory information
     into the agent's final output memory information.

    Args:
        memory (Memory): The current memory instance.
        agent_input (dict): Agent input object.
        content (str): The content of the current memory message.
        source (str): The source of the current memory message.
        memory_messages (List[Message]): The historical memory messages.
    Returns:
        list[Message]: The assembled final output memory information.
    """
    cur_memory_message = Message(content=content, source=source)
    if memory:
        # add the current memory message to the memory instance.
        memory.add([cur_memory_message], **agent_input)
    if memory_messages is None:
        memory_messages = []
    memory_messages.append(cur_memory_message)
    return memory_messages



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/13 15:22
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: prompt.py
"""Prompt base module."""
import re
from typing import Optional

from langchain_core.prompts import PromptTemplate

from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.component_configer.configers.prompt_configer import PromptConfiger
from agentuniverse.base.util.prompt_util import generate_template
from agentuniverse.prompt.prompt_model import AgentPromptModel


class Prompt(ComponentBase):
    """Prompt class."""

    prompt_version: Optional[str] = None
    prompt_template: Optional[str] = None
    input_variables: Optional[list[str]] = None

    def __init__(self, **kwargs):
        super().__init__(component_type=ComponentEnum.PROMPT, **kwargs)

    def as_langchain(self):
        """Convert the prompt template into a LangChain prompt template.

        Returns:
            PromptTemplate: The prompt template.
        """
        return PromptTemplate(template=self.prompt_template,
                              input_variables=self.input_variables)

    def build_prompt(self, agent_prompt_model: AgentPromptModel, prompt_assemble_order: list[str]) -> 'Prompt':
        """Build the prompt class.

        Args:
            agent_prompt_model (AgentPromptModel): The user agent prompt model.
            prompt_assemble_order (list[str]): The prompt assemble ordered list.

        Returns:
            Prompt: The prompt object.
        """
        self.prompt_template = generate_template(agent_prompt_model, prompt_assemble_order)
        self.input_variables = re.findall(r'\{(.*?)}', self.prompt_template)
        return self

    def get_instance_code(self) -> str:
        """Return the prompt version of the current prompt."""
        return self.prompt_version

    def initialize_by_component_configer(self, component_configer: PromptConfiger) -> 'Prompt':
        """Initialize the prompt by the PromptConfiger object.

        Args:
            component_configer(PromptConfiger): the PromptConfiger object
        Returns:
            Prompt: the prompt object
        """
        prompt_values = []
        for k, v in component_configer.configer.value.items():
            # ignore metadata values
            if k == 'metadata':
                continue
            self.__dict__[k] = v
            # splice the values in the prompt configer
            prompt_values.append(str(v))

        if component_configer.metadata_version:
            self.prompt_version = component_configer.metadata_version

        self.prompt_template = '\n'.join(prompt_values)

        self.input_variables = re.findall(r'\{(.*?)}', self.prompt_template)
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/22 14:45
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: prompt_configer.py
from typing import Optional

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer


class PromptConfiger(ComponentConfiger):
    """The PromptConfiger class, which is used to load and manage the Prompt configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the PromptConfiger."""
        super().__init__(configer)
        self.__metadata_version: Optional[str] = None

    @property
    def metadata_version(self) -> Optional[str]:
        """Return prompt version of the Prompt."""
        return self.__metadata_version

    def load(self) -> 'PromptConfiger':
        """Load the configuration by the Configer object.
        Returns:
            PromptConfiger: the PromptConfiger object
        """
        return self.load_by_configer(self.__configer)

    def load_by_configer(self, configer: Configer) -> 'PromptConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            PromptConfiger: the PromptConfiger object
        """
        super().load_by_configer(configer)
        try:
            self.__metadata_version = configer.value.get('metadata').get('version')
            # set the prompt default module and class
            if self.metadata_module is None:
                self.metadata_module = 'agentuniverse.prompt.prompt'
            if self.metadata_class is None:
                self.metadata_class = 'Prompt'
        except Exception as e:
            raise Exception(f"Failed to parse the Prompt configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/8 15:23
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: temp.py
import jieba
from collections import defaultdict, Counter
import math


# 构建倒排索引
def build_inverted_index(corpus):
    inverted_index = defaultdict(list)
    for idx, document in enumerate(corpus):
        words = jieba.lcut(document)
        for word in set(words):
            inverted_index[word].append(idx)
    return inverted_index


# 计算 BM25
def compute_bm25(query, doc_id, corpus, inverted_index, k1=1.5, b=0.75):
    query_words = jieba.lcut(query)
    doc_words = jieba.lcut(corpus[doc_id])
    doc_length = len(doc_words)
    avg_doc_length = sum(len(jieba.lcut(doc)) for doc in corpus) / len(corpus)
    bm25_score = 0

    doc_counter = Counter(doc_words)

    for term in query_words:
        if term in inverted_index:
            term_freq = doc_counter[term]
            num_docs_with_term = len(inverted_index[term])
            idf = math.log((len(corpus) - num_docs_with_term + 0.5) / (
                        num_docs_with_term + 0.5) + 1)
            bm25_term_score = idf * (term_freq * (k1 + 1)) / (
                        term_freq + k1 * (
                            1 - b + b * (doc_length / avg_doc_length)))
            bm25_score += bm25_term_score

    return bm25_score


# 示例数据
corpus = [
    "自然语言处理是人工智能的一个重要方向",
    "我喜欢学习新的技术",
    "自然语言处理包括很多技术，比如分词、词性标注、命名实体识别等"
]

# 构建倒排索引
inverted_index = build_inverted_index(corpus)

# 查询
query = "自然语言处理"
scores = {idx: compute_bm25(query, idx, corpus, inverted_index) for idx in
          range(len(corpus))}

# 排序并输出最相关的文档
sorted_docs = sorted(scores.items(), key=lambda item: item[1], reverse=True)
for doc_id, score in sorted_docs:
    print(f"文档ID: {doc_id}, BM25分数: {score}, 内容: {corpus[doc_id]}")

if __name__ == "__main__":
    print("test")


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 15:44
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: knowledge.py
import os
import re
from typing import Optional, Dict, List, Any
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED

from langchain_core.utils.json import parse_json_markdown
from langchain.tools import Tool as LangchainTool

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.store.store_manager import StoreManager
from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import DocProcessor
from agentuniverse.agent.action.knowledge.doc_processor.doc_processor_manager import DocProcessorManager
from agentuniverse.agent.action.knowledge.query_paraphraser.query_paraphraser import QueryParaphraser
from agentuniverse.agent.action.knowledge.query_paraphraser.query_paraphraser_manager import QueryParaphraserManager
from agentuniverse.agent.action.knowledge.rag_router.rag_router_manager import RagRouterManager
from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.reader.reader_manager import ReaderManager
from agentuniverse.base.annotation.trace import trace_knowledge
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.util.logging.logging_util import LOGGER


class Knowledge(ComponentBase):
    """
    The basic class for the knowledge model.

    Attributes:
        name (str): The name of the knowledge.

        description (str): The description of the knowledge.

        stores (List[str]): The stores for the knowledge, which are used to store knowledge
            and provide retrieval capabilities, such as ChromaDB store or Redis Store.

        query_paraphrasers (List[str]): Query paraphrasers used to paraphrase the original query string,
            such as extracting keywords and splitting into sub-queries.

        insert_processors (List[str]): DocProcessors used in the knowledge insertion step,
            such as text splitter and text cleaner.

        rag_router (str): RAG router used to decide which stores to use in
            the RAG step.

        post_processors (List[str]): DocProcessors used in the RAG step to process retrieved
            documents, such as reranking and filtering.

        readers (Dict[str, str]): The readers of the knowledge, which are used to load data and generate knowledge.
            Each reader refers to a specific file type.

        insert_executor (ThreadPoolExecutor): Used for performing insert and search
        operations concurrently in multiple stores.

        ext_info (Optional[Dict]): The extended information of the knowledge.
    """
    class Config:
        arbitrary_types_allowed = True

    name: str = ""
    description: Optional[str] = None
    stores: List[str] = []
    query_paraphrasers: List[str] = []
    insert_processors: List[str] = []
    update_processors: List[str] = []
    rag_router: str = "base_router"
    post_processors: List[str] = []
    readers: Dict[str, str] = dict()
    insert_executor: Optional[ThreadPoolExecutor] = None
    query_executor: Optional[ThreadPoolExecutor] = None
    ext_info: Optional[Dict] = None

    def __init__(self, **kwargs):
        super().__init__(component_type=ComponentEnum.KNOWLEDGE, **kwargs)
        self.insert_executor = ThreadPoolExecutor(
            max_workers=5,
            thread_name_prefix="Knowledge store"
        )
        self.query_executor = ThreadPoolExecutor(
            max_workers=10,
            thread_name_prefix="Knowledge query"
        )

    def _load_data(self,  *args: Any, **kwargs: Any) -> List[Document]:
        # check if source is a local file or remote url
        if kwargs.get("source_path"):
            source_path = kwargs.get("source_path")
        else:
            raise Exception("No file to load.")
        url_pattern = re.compile(
            r'^(https?:\/\/)?' 
            r'((([a-zA-Z0-9]{1,256}\.[a-zA-Z0-9]{1,6})|'
            r'(\d{1,3}\.){3}\d{1,3})'
            r'(:\d{1,5})?)'
            r'(\/[a-zA-Z0-9@:%._\+~#=]*)*\/?'
            r'(\?[a-zA-Z0-9@:%._\+~#&//=]*)?$'
        )

        if url_pattern.match(source_path):
            source_type = "url"
        elif os.path.isfile(source_path):
            source_type = os.path.splitext(source_path)[1][1:]
        else:
            raise Exception(f"Knowledge load data error: Unknown source type:{source_path}")
        if source_type in self.readers:
            reader = ReaderManager().get_instance_obj(self.readers[source_type])
        else:
            reader = ReaderManager().get_file_default_reader(source_type)
        return reader.load_data(source_path)

    def _insert_process(self, origin_docs: List[Document]) -> List[Document]:
        for _processor_code in self.insert_processors:
            doc_processor: DocProcessor = DocProcessorManager().get_instance_obj(_processor_code)
            origin_docs = doc_processor.process_docs(origin_docs)
        return origin_docs

    def _update_process(self, origin_docs: List[Document]) -> List[Document]:
        for _processor_code in self.update_processors:
            doc_processor: DocProcessor = DocProcessorManager().get_instance_obj(_processor_code)
            origin_docs = doc_processor.process_docs(origin_docs)
        return origin_docs

    def _rag_post_process(self, origin_docs: List[Document], query: Query):
        for _processor_code in self.post_processors:
            doc_processor: DocProcessor = DocProcessorManager().get_instance_obj(_processor_code)
            origin_docs = doc_processor.process_docs(origin_docs, query=query)
        return origin_docs

    def _paraphrase_query(self, origin_query: Query) -> Query:
        for _paraphraser_code in self.query_paraphrasers:
            query_paraphraser: QueryParaphraser = QueryParaphraserManager().get_instance_obj(
                _paraphraser_code)
            origin_query = query_paraphraser.query_paraphrase(origin_query)
        return origin_query

    def insert_knowledge(self, **kwargs) -> None:
        """Insert the knowledge.

        Load data by the reader and insert the documents into the store.
        """
        document_list: List[Document] = self._load_data(**kwargs)
        document_list = self._insert_process(document_list)
        futures = []
        if "stores" in kwargs:
            stores = kwargs["stores"]
        else:
            stores = self.stores
        for _store_code in stores:
            futures.append(
                self.insert_executor.submit(
                    StoreManager().get_instance_obj(_store_code).insert_document,
                    document_list))
        wait(futures, return_when=ALL_COMPLETED)
        for future in futures:
            try:
                future.result()
            except Exception as e:
                LOGGER.error(f"Exception occurred in knowledge insert: {e}")
        LOGGER.info("Knowledge insert complete.")

    def update_knowledge(self, **kwargs) -> None:
        """Update the knowledge.

        Load data by the reader and update the documents into the store.
        """
        document_list: List[Document] = self._load_data(**kwargs)
        document_list = self._update_process(document_list)
        futures = []
        if "stores" in kwargs:
            stores = kwargs["stores"]
        else:
            stores = self.stores
        for _store_code in stores:
            futures.append(
                self.insert_executor.submit(
                    StoreManager().get_instance_obj(_store_code).update_document,
                    document_list))
        wait(futures, return_when=ALL_COMPLETED)
        for future in futures:
            try:
                future.result()
            except Exception as e:
                LOGGER.error(f"Exception occurred in knowledge update: {e}")
        LOGGER.info("Knowledge update complete.")

    def _route_rag(self, query: Query):
        return RagRouterManager().get_instance_obj(self.rag_router).rag_route(query, self.stores)

    @trace_knowledge
    def query_knowledge(self, **kwargs) -> List[Document]:
        """Query the knowledge.

        Query documents from the store and return the results.
        """
        query = Query(**kwargs)
        query = self._paraphrase_query(query)
        query_tasks = self._route_rag(query)

        futures = []
        for query_task in query_tasks:
            futures.append(
                self.query_executor.submit(
                    StoreManager().get_instance_obj(query_task[1]).query,
                    query_task[0]))
        wait(futures, return_when=ALL_COMPLETED)
        retrieved_docs = {}
        for future in futures:
            try:
                task_result = future.result()
                for _doc in task_result:
                    if _doc.id not in retrieved_docs:
                        retrieved_docs[_doc.id] = _doc
            except Exception as e:
                LOGGER.error(f"Exception occurred in knowledge query: {e}")
        retrieved_docs = list(retrieved_docs.values())
        retrieved_docs = self._rag_post_process(retrieved_docs, query)
        return retrieved_docs

    def to_llm(self, retrieved_docs: List[Document]) -> Any:
        """Transfer list docs to llm input"""
        retrieved_texts = [doc.text for doc in retrieved_docs]
        return "\n=========================================\n".join(retrieved_texts)

    def _initialize_by_component_configer(self,
                                          knowledge_configer: ComponentConfiger) \
            -> 'Knowledge':
        """Initialize the reader by the ComponentConfiger object.

        Args:
            reader_configer(ComponentConfiger): A configer contains reader
            basic info.
        Returns:
            Reader: A reader instance.
        """
        if knowledge_configer.name:
            self.name = knowledge_configer.name
        if knowledge_configer.description:
            self.description = knowledge_configer.description
        if hasattr(knowledge_configer, "stores"):
            self.stores = knowledge_configer.stores
        if hasattr(knowledge_configer, "query_paraphrasers"):
            self.query_paraphrasers = knowledge_configer.query_paraphrasers
        if hasattr(knowledge_configer, "insert_processors"):
            self.insert_processors = knowledge_configer.insert_processors
        if hasattr(knowledge_configer, "update_processors"):
            self.update_processors = knowledge_configer.update_processors
        if hasattr(knowledge_configer, "rag_router"):
            self.rag_router = knowledge_configer.rag_router
        if hasattr(knowledge_configer, "post_processors"):
            self.post_processors = knowledge_configer.post_processors
        if hasattr(knowledge_configer, "readers"):
            self.readers = knowledge_configer.readers
        return self

    def langchain_query(self, query: str) -> str:
        """Query the knowledge using LangChain.

        Query documents from the store and return the results.
        """
        parse_query = parse_json_markdown(query)
        knowledge = self.query_knowledge(**parse_query)
        return "This is Query Result:\n"+self.to_llm(knowledge)

    def as_langchain_tool(self) -> LangchainTool:
        """Convert the Knowledge object to a LangChain tool.

        Returns:
            Any: the LangChain tool object
        """
        args_description = """
        This is a knowledge base tool, which stores the content you may need. To use this tool, you need to give a json string with the following format:
        ```json
        {
            "query_str": "<your query here>",
            "similarity_top_k": <number of results to return>,
        }
        ```
        """
        return LangchainTool(
            name=self.name,
            description=self.description + args_description,
            func=self.langchain_query,
        )



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/27 15:53
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: plugin_dto.py
from typing import Optional, List

from pydantic import BaseModel, Field

from agentuniverse_product.service.model.tool_dto import ToolDTO


class PluginDTO(BaseModel):
    id: str = Field(description="ID")
    nickname: Optional[str] = Field(description="plugin nickname", default="")
    avatar: Optional[str] = Field(description="plugin avatar path", default="")
    description: Optional[str] = Field(description="plugin description", default="")
    toolset: Optional[List[ToolDTO]] = Field(description="plugin toolset", default=[])
    openapi_desc: Optional[str] = Field(description="plugin openapi schema", default="")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 15:18
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: request_library.py
import datetime

from sqlalchemy import JSON, Integer, String, DateTime, Text, Column
from sqlalchemy import select
from sqlalchemy.orm import declarative_base

from .entity.request_do import RequestDO
from agentuniverse.base.util.system_util import get_project_root_path
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.config.component_configer.configers.sqldb_wrapper_config import SQLDBWrapperConfiger
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.database.sqldb_wrapper import SQLDBWrapper
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager

REQUEST_TABLE_NAME = 'request_task'
Base = declarative_base()


class RequestORM(Base):
    """SQLAlchemy ORM Model for RequestDO."""
    __tablename__ = REQUEST_TABLE_NAME
    id = Column(Integer, primary_key=True, autoincrement=True)
    request_id = Column(String(20), nullable=False)
    query = Column(Text)
    session_id = Column(String(50))
    state = Column(String(20))
    result = Column(JSON)
    steps = Column(JSON)
    additional_args = Column(JSON)
    gmt_create = Column(DateTime, default=datetime.datetime.now)
    gmt_modified = Column(DateTime, default=datetime.datetime.now,
                          onupdate=datetime.datetime.now)


@singleton
class RequestLibrary:
    def __init__(self, configer: Configer = None):
        """Init the database connection. Use uri in config file or use sqlite
        as default database."""
        system_db_uri = None
        if Configer:
            system_db_uri = configer.get('DB', {}).get('system_db_uri')
            if not system_db_uri:
                system_db_uri = configer.get('DB', {}).get('mysql_uri')
        if system_db_uri and system_db_uri.strip():
            pass
        else:
            db_path = get_project_root_path() / 'intelligence' / 'db' / 'agent_universe.db'
            db_path.parent.mkdir(parents=True, exist_ok=True)
            system_db_uri = f'sqlite:///{db_path}'

        self.session = None
        # create a sqldb_wrapper_instance
        _configer = SQLDBWrapperConfiger()
        _configer.db_uri = system_db_uri
        self.sqldb_wrapper = SQLDBWrapper(
            name="__system_db__",
            db_wrapper_configer=_configer
        )
        SQLDBWrapperManager().register(self.sqldb_wrapper.get_instance_code(),
                                       self.sqldb_wrapper)

    def __init_request_table(self):
        with self.sqldb_wrapper.sql_database._engine.connect() as conn:
            if not conn.dialect.has_table(conn, REQUEST_TABLE_NAME):
                Base.metadata.create_all(self.sqldb_wrapper.sql_database._engine)

    def get_session(self):
        if not self.session:
            self.__init_request_table()
            self.session = self.sqldb_wrapper.get_session()
        return self.session()

    def query_request_by_request_id(self, request_id: str) -> RequestDO | None:
        """Get a RequestDO with given request_id.

        Args:
            request_id(`str`): The unique request id of request task.

        Return:
            The target RequestDO or none when no such data.
        """
        session = self.get_session()
        try:
            result = session.execute(
                select(RequestORM).where(RequestORM.request_id == request_id)
            ).scalars().first()
            if not result:
                return None
            return self.__request_orm_to_do(result)
        finally:
            session.close()

    def add_request(self, request_do: RequestDO) -> int:
        """Add the given RequestDO to database.

        Args:
            request_do(`RequestDO`): A new RequestDO to be added.

        Return:
            A int stands unique data id in table.
        """
        session = self.get_session()
        try:
            request_orm = RequestORM(**request_do.model_dump())
            session.add(request_orm)
            session.commit()
            return request_orm.id
        finally:
            session.close()

    def update_request(self, request_do: RequestDO):
        """Update the request data with same request id as the given
        RequestDO."""
        session = self.get_session()
        try:
            db_request_do = session.query(RequestORM).filter(
                RequestORM.request_id == request_do.request_id).first()
            if db_request_do:
                update_data = request_do.model_dump(exclude_unset=True)
                for key, value in update_data.items():
                    setattr(db_request_do, key, value)
                session.commit()
                session.refresh(db_request_do)
        finally:
            session.close()

    def update_gmt_modified(self, request_id: str):
        """Update the request task latest active time."""
        session = self.get_session()
        try:
            db_request_do = session.query(RequestORM).filter(
                RequestORM.request_id == request_id).first()
            if db_request_do:
                setattr(db_request_do, "gmt_modified", datetime.datetime.now())
                session.commit()
                session.refresh(db_request_do)
        finally:
            session.close()

    def __request_orm_to_do(self, request_orm: RequestORM) -> RequestDO:
        """Transfer a RequestORM to RequestDO."""
        request_obj = RequestDO(
            request_id='',
            session_id="",
            query='',
            state='',
            result=dict(),
            steps=[],
            additional_args=dict(),
            gmt_create=datetime.datetime.now(),
            gmt_modified=datetime.datetime.now(),
        )
        for column in request_orm.__table__.columns:
            setattr(request_obj, column.name,
                    getattr(request_orm, column.name))
        return request_obj



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/21 20:52
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_service.py
import os
import random

from agentuniverse.workflow.workflow import Workflow
from agentuniverse.workflow.workflow_manager import WorkflowManager
from agentuniverse_product.base.util.yaml_util import write_yaml_file, update_nested_yaml_value
from agentuniverse_product.service.model.workflow_dto import WorkflowDTO
from agentuniverse_product.service.util.common_util import get_core_path
from agentuniverse_product.service.util.workflow_util import register_workflow


class WorkflowService:
    """Workflow Service for aU-product."""

    @staticmethod
    def create_workflow(workflow_dto: WorkflowDTO) -> str:
        """Creates a workflow and returns its ID.

        Args:
            workflow_dto (WorkflowDTO): The workflow DTO containing the workflow attributes.

        Returns:
            str: The ID of the created workflow.
        """
        workflow_id = f"workflow_{random.randint(100, 999)}"
        workflow_config = {
            'id': workflow_id,
            'name': workflow_dto.name,
            'description': workflow_dto.description,
            'graph': workflow_dto.graph,
            'metadata': {
                'class': 'Workflow',
                'module': 'agentuniverse.workflow.workflow',
                'type': 'WORKFLOW'
            }
        }
        # write workflow YAML file
        path = get_core_path()
        workflow_file_path = path / 'workflow' / f"{workflow_id}.yaml" if path \
            else os.path.join('..', '..', 'platform', 'difizen', 'workflow', f"{workflow_id}.yaml")
        write_yaml_file(str(workflow_file_path), workflow_config)
        # register workflow instance
        register_workflow(str(workflow_file_path))
        return workflow_id

    @staticmethod
    def update_workflow(workflow_dto: WorkflowDTO) -> str:
        """Update workflow attributes.

        Args:
            workflow_dto (WorkflowDTO): The workflow DTO containing the workflow attributes.

        Returns:
            str: The ID of the updated workflow.
        """
        if workflow_dto.id is None:
            raise ValueError("workflow_id is None.")
        workflow: Workflow = WorkflowManager().get_instance_obj(component_instance_name=workflow_dto.id)
        if workflow is None:
            raise ValueError("The workflow instance corresponding to the workflow_id cannot be found.")
        workflow_update_config = {
            'name': workflow_dto.name,
            'description': workflow_dto.description,
            'graph_config': workflow_dto.graph
        }
        # Update the workflow attributes if they are provided
        updated_fields = {key: value for key, value in workflow_update_config.items() if value is not None}

        # set workflow attributes
        for key, value in updated_fields.items():
            setattr(workflow, key, value)

        if updated_fields:
            graph_config = updated_fields.pop('graph_config', None)
            updated_fields['graph'] = graph_config
            update_nested_yaml_value(workflow.component_config_path, updated_fields)
        return workflow_dto.id

    @staticmethod
    def get_workflow_detail(id: str) -> WorkflowDTO | None:
        """Get workflow detail by workflow id.

        Args:
            id (str): The ID of the workflow.

        Returns:
            WorkflowDTO | None: WorkflowDTO or None.
        """
        workflow: Workflow = WorkflowManager().get_instance_obj(component_instance_name=id)
        if workflow is None:
            raise ValueError("The workflow instance corresponding to the workflow_id cannot be found.")
        return WorkflowDTO(id=workflow.id, name=workflow.name, description=workflow.description,
                           graph=workflow.graph_config)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/10 19:10
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: chroma_memory_storage.py
import uuid
from datetime import datetime
from urllib.parse import urlparse
from typing import Optional, List, Any

import chromadb
from pydantic import SkipValidation
from chromadb.config import Settings
from chromadb.api.models.Collection import Collection

from agentuniverse.agent.action.knowledge.embedding.embedding_manager import EmbeddingManager
from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class ChromaMemoryStorage(MemoryStorage):
    """The chroma memory storage class.

    Attributes:
        collection_name (str): The name of the ChromaDB collection.
        persist_path (str): The path to persist the collection.
        embedding_model (str): The name of the embedding model instance to use.
        _collection (Collection): The collection object.
    """
    collection_name: Optional[str] = 'memory'
    persist_path: Optional[str] = None
    embedding_model: Optional[str] = None
    _collection: SkipValidation[Collection] = None

    def _initialize_by_component_configer(self,
                                          memory_storage_config: ComponentConfiger) -> 'ChromaMemoryStorage':
        """Initialize the ChromaMemoryStorage by the ComponentConfiger object.

        Args:
            memory_storage_config(ComponentConfiger): A configer contains chroma_memory_storage basic info.
        Returns:
            ChromaMemoryStorage: A ChromaMemoryStorage instance.
        """
        super()._initialize_by_component_configer(memory_storage_config)
        if getattr(memory_storage_config, 'collection_name', None):
            self.collection_name = memory_storage_config.collection_name
        if getattr(memory_storage_config, 'persist_path', None):
            self.persist_path = memory_storage_config.persist_path
        if getattr(memory_storage_config, 'embedding_model', None):
            self.embedding_model = memory_storage_config.embedding_model
        return self

    def _init_collection(self) -> Any:
        """Initialize the ChromaDB collection."""
        if self.persist_path.startswith('http') or self.persist_path.startswith('https'):
            parsed_url = urlparse(self.persist_path)
            settings = Settings(
                chroma_api_impl="chromadb.api.fastapi.FastAPI",
                chroma_server_host=parsed_url.hostname,
                chroma_server_http_port=str(parsed_url.port)
            )
        else:
            settings = Settings(
                is_persistent=True,
                persist_directory=self.persist_path
            )
        client = chromadb.Client(settings)
        self._collection = client.get_or_create_collection(name=self.collection_name)
        return client

    def delete(self, session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Delete the memory from the database.

        Args:
            session_id (str): The session id of the memory to delete.
            agent_id (str): The agent id of the memory to delete.
        """
        if self._collection is None:
            self._init_collection()
        filters = {}
        if session_id is None and agent_id is None:
            return
        if session_id is not None:
            filters['session_id'] = session_id
        if agent_id is not None:
            filters['agent_id'] = agent_id
        self._collection.delete(where=filters)

    def add(self, message_list: List[Message], session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Add messages to the memory db.

        Args:
            message_list (List[Message]): The list of messages to add.
            session_id (str): The session id of the memory to add.
            agent_id (str): The agent id of the memory to add.
        """
        if self._collection is None:
            self._init_collection()
        if not message_list:
            return
        metadata = {'gmt_created': datetime.now().isoformat()}
        if session_id:
            metadata['session_id'] = session_id
        if agent_id:
            metadata['agent_id'] = agent_id
        for message in message_list:
            embedding = []
            if self.embedding_model:
                embedding = EmbeddingManager().get_instance_obj(
                    self.embedding_model
                ).get_embeddings([message.content])[0]
            if message.source:
                metadata['source'] = message.source
            metadata['type'] = message.type if message.type else ''
            self._collection.add(
                ids=[message.id if message.id else str(uuid.uuid4())],
                documents=[message.content],
                metadatas=[metadata],
                embeddings=[embedding] if len(embedding) > 0 else None,
            )

    def get(self, session_id: str = None, agent_id: str = None, top_k=10, input: str = '', source: str = None, **kwargs) -> \
            List[Message]:
        """Get messages from the memory db.

        Args:
            session_id (str): The session id of the memory to get.
            agent_id (str): The agent id of the memory to get.
            top_k (int): The number of messages to return.
            input (str): The input text to search for in the memory.
            source (str): The source of the message to get.
        Returns:
            List[Message]: A list of messages retrieved from the memory db.
        """
        if self._collection is None:
            self._init_collection()
        filters = {"$and": []}
        if session_id:
            filters["$and"].append({'session_id': session_id})
        if agent_id:
            filters["$and"].append({'agent_id': agent_id})
        if source:
            filters["$and"].append({'source': source})
        if len(filters["$and"]) < 2:
            filters = filters["$and"][0] if len(filters["$and"]) == 1 else {}
        if input:
            embedding = []
            if self.embedding_model:
                embedding = EmbeddingManager().get_instance_obj(
                    self.embedding_model
                ).get_embeddings([input])[0]
            if len(embedding) > 0:
                results = self._collection.query(
                    query_embeddings=embedding, where=filters, n_results=top_k
                )
            else:
                results = self._collection.query(query_texts=[input], where=filters, n_results=top_k)
            messages = self.to_messages(result=results)
            messages.reverse()
            return messages
        else:
            results = self._collection.get(where=filters)
            messages = self.to_messages(result=results, sort_by_time=True)
            return messages[-top_k:]

    def to_messages(self, result: dict, sort_by_time: bool = False) -> List[Message]:
        """Convert the result from ChromaDB to a list of aU messages.

        Args:
            result (dict): The result from ChromaDB.
            sort_by_time (bool): Whether to sort the messages by time.
        Returns:
            List[Message]: A list of aU messages.
        """
        message_list = []
        if not result or not result['ids']:
            return message_list
        try:
            if self.is_nested_list(result['ids']):
                metadatas = result.get('metadatas', [[]])
                documents = result.get('documents', [[]])
                ids = result.get('ids', [[]])
                message_list = [
                    Message(
                        id=ids[0][i],
                        content=documents[0][i],
                        metadata=metadatas[0][i] if metadatas[0] else None,
                        source=metadatas[0][i].get('source', None) if metadatas[0] else None,
                        type=metadatas[0][i].get('type', '') if metadatas[0] else ''
                    )
                    for i in range(len(result['ids'][0]))
                ]
            else:
                metadatas = result.get('metadatas', [])
                documents = result.get('documents', [])
                ids = result.get('ids', [])
                message_list = [
                    Message(
                        id=ids[i],
                        content=documents[i],
                        metadata=metadatas[i] if metadatas[i] else None,
                        source=metadatas[i].get('source', None) if metadatas[i] else None,
                        type=metadatas[i].get('type', '') if metadatas[i] else ''
                    )
                    for i in range(len(result['ids']))
                ]
            if sort_by_time:
                # order by gmt_created asc
                message_list = sorted(
                    message_list,
                    key=lambda msg: msg.metadata.get('gmt_created', ''),
                )
        except Exception as e:
            print('ChromaMemory.to_messages failed, exception= ' + str(e))
        return message_list

    @staticmethod
    def is_nested_list(variable: List) -> bool:
        return isinstance(variable, list) and len(variable) > 0 and isinstance(variable[0], list)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/26 11:02
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: knowledge_dto.py
from typing import Optional

from pydantic import BaseModel, Field


class KnowledgeDTO(BaseModel):
    id: Optional[str] = Field(description="ID", default="")
    nickname: Optional[str] = Field(description="knowledge nickname", default="")
    description: Optional[str] = Field(description="knowledge description", default="")
    avatar: Optional[str] = Field(description="knowledge avatar path", default="")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/24 21:56
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: nl2api_agent_template.py
from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger


class Nl2ApiAgentTemplate(AgentTemplate):

    def input_keys(self) -> list[str]:
        return ['input']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        agent_input['tools'] = self.build_tools_context()
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        return {**agent_result, 'output': agent_result['output']}

    def build_tools_context(self) -> str:
        tools_context = ''
        if self.tool_names:
            for tool_name in self.tool_names:
                tool: Tool = ToolManager().get_instance_obj(tool_name)
                tools_context += f"tool name: {tool.name}, tool description: {tool.description}\n"
        return tools_context

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'Nl2ApiAgentTemplate':
        super().initialize_by_component_configer(component_configer)
        self.prompt_version = self.agent_model.profile.get('prompt_version', 'default_nl2api_agent.cn')
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/25 15:06
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: react_agent_template.py
from typing import Sequence, Optional, Union, List

from langchain.agents.format_scratchpad import format_log_to_str
from langchain.agents.output_parsers import ReActSingleInputOutputParser
from langchain.agents import AgentExecutor, AgentOutputParser
from langchain.tools import Tool as LangchainTool
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import BasePromptTemplate
from langchain_core.runnables import RunnableConfig, RunnablePassthrough, Runnable
from langchain_core.tools import BaseTool, ToolsRenderer, render_text_description

from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent import Agent
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.plan.planner.react_planner.stream_callback import StreamOutPutCallbackHandler
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.prompt import Prompt


class ReActAgentTemplate(AgentTemplate):
    agent_names: Optional[list[str]] = None
    stop_sequence: Optional[list[str]] = None
    max_iterations: Optional[int] = None

    def input_keys(self) -> list[str]:
        return ['input']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        tools_context = self.build_tools_context()
        agent_input['tools'] = tools_context[0]
        agent_input['tool_names'] = tools_context[1]
        agent_input['agent_scratchpad'] = ''
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        return {**agent_result, 'output': agent_result['output']}

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM, prompt: Prompt,
                           **kwargs) -> dict:
        assemble_memory_input(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        lc_tools: List[LangchainTool] = self._convert_to_langchain_tool()
        agent = self.create_react_agent(llm.as_langchain(), lc_tools, prompt.as_langchain(),
                                        stop_sequence=self.stop_sequence,
                                        bind_params=self.agent_model.llm_params())
        agent_executor = AgentExecutor(agent=agent, tools=lc_tools,
                                       verbose=True,
                                       handle_parsing_errors=True,
                                       max_iterations=self.max_iterations)
        res = agent_executor.invoke(input=agent_input, memory=memory.as_langchain() if memory else None,
                                    chat_history=agent_input.get(memory.memory_key) if memory else '',
                                    config=self._get_run_config(input_object))
        assemble_memory_output(memory=memory,
                               agent_input=agent_input,
                               content=f"Human: {agent_input.get('input')}, AI: {res.get('output')}")
        return res

    async def customized_async_execute(self, input_object: InputObject, agent_input: dict, memory: Memory,
                                       llm: LLM, prompt: Prompt, **kwargs) -> dict:
        assemble_memory_input(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        lc_tools: List[LangchainTool] = self._convert_to_langchain_tool()
        agent = self.create_react_agent(llm.as_langchain(), lc_tools, prompt.as_langchain(),
                                        stop_sequence=self.stop_sequence,
                                        bind_params=self.agent_model.llm_params())
        agent_executor = AgentExecutor(agent=agent, tools=lc_tools,
                                       verbose=True,
                                       handle_parsing_errors=True,
                                       max_iterations=self.max_iterations)
        res = await agent_executor.ainvoke(input=agent_input, memory=memory.as_langchain() if memory else None,
                                           chat_history=agent_input.get(memory.memory_key) if memory else '',
                                           config=self._get_run_config(input_object))
        assemble_memory_output(memory=memory,
                               agent_input=agent_input,
                               content=f"Human: {agent_input.get('input')}, AI: {res.get('output')}")
        return res

    def create_react_agent(
            self,
            llm: BaseLanguageModel,
            tools: Sequence[BaseTool],
            prompt: BasePromptTemplate,
            output_parser: Optional[AgentOutputParser] = None,
            tools_renderer: ToolsRenderer = render_text_description,
            *,
            stop_sequence: Union[bool, List[str]] = True,
            bind_params: Optional[dict],
    ) -> Runnable:
        missing_vars = {"tools", "tool_names", "agent_scratchpad"}.difference(
            prompt.input_variables + list(prompt.partial_variables)
        )
        if missing_vars:
            raise ValueError(f"Prompt missing required variables: {missing_vars}")

        prompt = prompt.partial(
            tools=tools_renderer(list(tools)),
            tool_names=", ".join([t.name for t in tools]),
        )
        if stop_sequence:
            stop = ["\nObservation"] if stop_sequence is True else stop_sequence
            llm_with_stop = llm.bind(stop=stop, **(bind_params or {}))
        else:
            llm_with_stop = llm.bind(**(bind_params or {}))
        output_parser = output_parser or ReActSingleInputOutputParser()
        agent = (
                RunnablePassthrough.assign(
                    agent_scratchpad=lambda x: format_log_to_str(x["intermediate_steps"]),
                )
                | prompt
                | llm_with_stop
                | output_parser
        )
        return agent

    def build_tools_context(self) -> tuple[str, str]:
        lc_tools: [LangchainTool] = self._convert_to_langchain_tool()
        tools_context = ''
        if not lc_tools:
            return '', ''
        tool_names = []
        for lc_tool in lc_tools:
            tools_context += f"tool name: {lc_tool.name}, tool description: {lc_tool.description}\n"
        return tools_context, "|".join(tool_names)

    def _convert_to_langchain_tool(self) -> list[LangchainTool]:
        lc_tools = []
        if self.tool_names:
            for tool_name in self.tool_names:
                tool: Tool = ToolManager().get_instance_obj(tool_name)
                lc_tools.append(tool.as_langchain())
        if self.knowledge_names:
            for knowledge_name in self.knowledge_names:
                knowledge: Knowledge = KnowledgeManager().get_instance_obj(knowledge_name)
                lc_tools.append(knowledge.as_langchain_tool())
        if self.agent_names:
            for agent_name in self.agent_names:
                agent: Agent = AgentManager().get_instance_obj(agent_name)
                lc_tools.append(agent.as_langchain_tool())
        return lc_tools

    def _get_run_config(self, input_object: InputObject) -> RunnableConfig:
        config = RunnableConfig()
        callbacks = []
        output_stream = input_object.get_data('output_stream')
        callbacks.append(StreamOutPutCallbackHandler(output_stream, agent_info=self.agent_model.info))
        config.setdefault("callbacks", callbacks)
        return config

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'ReActAgentTemplate':
        super().initialize_by_component_configer(component_configer)
        self.prompt_version = self.agent_model.profile.get('prompt_version', 'default_react_agent.cn')
        self.stop_sequence = self.agent_model.profile.get('stop_sequence')
        self.max_iterations = self.agent_model.profile.get('max_iterations', 10)
        self.agent_names = self.agent_model.action.get('agent', [])
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/13 15:39
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: output_object.py
import json


class OutputObject(object):
    def __init__(self, params: dict):
        self.__params = params
        for k, v in params.items():
            self.__dict__[k] = v

    def to_dict(self):
        return self.__params

    def to_json_str(self):
        return json.dumps(self.__params, ensure_ascii=False)

    def get_data(self, key, default=None):
        return self.__params.get(key, default)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/13 11:20
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: nlu_rag_route_agent.py
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.template.rag_agent_template import RagAgentTemplate
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager


class NluRagRouteAgent(RagAgentTemplate):
    """Nlu Rag Route Agent class."""

    def input_keys(self) -> list[str]:
        """Return the input keys of the Agent."""
        return ['query', 'store_info', 'store_amount']

    def output_keys(self) -> list[str]:
        """Return the output keys of the Agent."""
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        """Agent parameter parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input preparsed by the agent.
        Returns:
            dict: agent input parsed from `input_object` by the user.
        """
        agent_input['query'] = input_object.get_data('query')
        agent_input['store_info'] = input_object.get_data('store_info')
        agent_input['store_amount'] = input_object.get_data('store_amount')
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        """Agent result parser.

        Args:
            agent_result(dict): The raw result of the agent.
        Returns:
            dict: The parsed result of the agent
        """
        return agent_result

    def process_llm(self, **kwargs) -> LLM:
        llm_name = self.agent_model.profile.get('llm_model', {}).get('name') or self.llm_name
        return LLMManager().get_instance_obj(llm_name)

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'NluRagRouteAgent':
        super().initialize_by_component_configer(component_configer)
        self.prompt_version = self.agent_model.profile.get('prompt_version', 'nlu_rag_route_prompt.cn')
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 14:29
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: config_manager.py
import os
import re
from typing import Optional, Callable
import tomli
import yaml

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.config.config_type_enum import ConfigTypeEnum

@singleton
class PlaceholderResolver:
    def __init__(self):
        self._resolvers = []
        self.register_resolver(r'${(.+?)}',
                                   lambda match: os.getenv(match.group(1),
                                                           match.group(0)))
    def register_resolver(self, pattern, func):
        """Register a new resolver with a regex pattern and its corresponding function."""
        self._resolvers.append((re.compile(pattern), func))

    def resolve(self, value):
        """Resolve placeholders in a given value based on registered resolvers."""
        if isinstance(value, dict):
            return {k: self.resolve(v) for k, v in value.items()}
        elif isinstance(value, list):
            return [self.resolve(item) for item in value]
        elif isinstance(value, str):
            for pattern, func in self._resolvers:
                value = pattern.sub(func, value)
            return value
        else:
            return value

class Configer(object):
    """Configger object, responsible for the configuration file load, update, etc."""

    # List of supported file formats
    __SUPPORTED_FILE_FORMATS = [ConfigTypeEnum.TOML.value, ConfigTypeEnum.YAML.value]

    def __init__(self,
                 path: str = None):
        """Initialize the ConfigManager
        Args:
            path(str): the path of the configuration file
        Returns:
            None
        """
        self.__path: str = path
        self.__value: dict = {}

    @property
    def path(self):
        """Return the path of the configuration file"""
        return self.__path

    @path.setter
    def path(self, path: str):
        """Set the path of the configuration file
        Args:
            path(str): the path of the configuration file
        Returns:
            None
        """
        self.__path = path

    @property
    def value(self):
        """Return the value of the configuration file"""
        return self.__value

    @value.setter
    def value(self, value: dict):
        """Set the value of the configuration file
        Args:
            value(dict): the value of the configuration file
        Returns:
            None
        """
        self.__value = value

    def load_by_path(self, path: str) -> 'Configer':
        """Load the configuration file by the given path
        Args:
            path(str): the path of the configuration file
        Returns:
            Configer: the Configer object
        """
        # Check the file format.
        file_format = path.split('.')[-1]
        if file_format not in self.__SUPPORTED_FILE_FORMATS:
            raise ValueError(f"Unsupported file format: {file_format}")

        # Choose the load method according to the file format.
        load_method = self.__choice_load_method(path)
        config_data = load_method(path)
        self.__value = config_data
        return self

    def load(self) -> 'Configer':
        """Load the configuration file

        Returns:
            Configer: the Configer object
        """
        return self.load_by_path(self.__path)

    def get(self, key: str, default=None) -> Optional[any]:
        """Return the value of the configuration file at the given key, or the default value if the key is not found
        Args:
            key(str): the key of the configuration file
            default(any): the default value
        Returns:
            Optional[any]: the value of the configuration file at the given key
        """
        return self.__value.get(key, default)

    def set(self, key: str, value):
        """Set the value of the configuration file at the given key.

        Args:
            key(str): the key of the configuration file
            value(any): the value of the configuration file
        Returns:
            None
        """
        self.__value[key] = value
        pass

    def to_dict(self) -> dict:
        """Return the dictionary representation of the configuration file.

        Returns:
            dict: the dictionary representation of the configuration file
        """
        return self.__value

    def __choice_load_method(self, path: str) -> Optional[Callable]:
        """Choose the load method according to the file format.

        Args:
            path(str): the path of the configuration file
        Returns:
            Optional[Callable]: the load method
        """
        # Define the regular expression and the corresponding method map.
        toml_re = re.compile(r'.*\.toml')
        yaml_re = re.compile(r'.*\.yaml')

        re_method_map = {
            toml_re: self.__load_toml_file,
            yaml_re: self.__load_yaml_file
        }
        for re_compile, method in re_method_map.items():
            if re_compile.search(path):
                return method

    @staticmethod
    def __load_toml_file(path: str) -> dict:
        """Load the toml file.

        Args:
            path(str): the path of the toml file
        Returns:
            dict: the value of the toml file
        """
        with open(path, 'rb') as f:
            config_data = tomli.load(f)
        config_data = PlaceholderResolver().resolve(config_data)
        return config_data

    @staticmethod
    def __load_yaml_file(path: str) -> dict:
        """Load the yaml file.

        Args:
            path(str): the path of the yaml file
        Returns:
            dict: the value of the yaml file
        """
        with open(path, 'r', encoding='utf-8') as stream:
            config_data = yaml.safe_load(stream)
        config_data = PlaceholderResolver().resolve(config_data)
        return config_data



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/6 16:00
# @Author  : wangyapei 
# @FileName: zhipu_openai_style_llm.py

from typing import Optional, Any, Union, Iterator, AsyncIterator

from pydantic import Field

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.openai_style_llm import OpenAIStyleLLM

ZHIPU_MAXCONTETNLENGTH = {
    "GLM-4-Plus": 128000,
    "GLM-4-0520": 128000,
    "GLM-4-AirX": 8000,
    "GLM-4-Air":  128000,
    "GLM-4-Long": 1000000,
    "GLM-4-Flash": 128000,
    "GLM-4": 128000,
}


class DefaultZhiPuLLM(OpenAIStyleLLM):
    """The agentUniverse default openai llm module.

    LLM parameters, such as name/description/model_name/max_tokens,
    are injected into this class by the default_openai_llm.yaml configuration.
    """

    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("ZHIPU_API_KEY"))
    organization: Optional[str] = Field(default_factory=lambda: get_from_env("ZHIPU_ORGANIZATION"))
    api_base: Optional[str] = Field(default_factory=lambda: get_from_env("ZHIPU_API_BASE"))
    proxy: Optional[str] = Field(default_factory=lambda: get_from_env("ZHIPU_PROXY"))

    @trace_llm
    def call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """ The call method of the LLM.

        Users can customize how the model interacts by overriding call method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return super().call(messages, **kwargs)

    @trace_llm
    async def acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """ The async call method of the LLM.

        Users can customize how the model interacts by overriding acall method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return await super().acall(messages, **kwargs)

    def max_context_length(self) -> int:
        """Max context length.

          The total length of input tokens and generated tokens is limited by the openai model's context length.
          """
        return ZHIPU_MAXCONTETNLENGTH.get(self.model_name, 128000)




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 16:31
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: chroma_store.py
from urllib.parse import urlparse
from typing import List, Any, Optional
from pydantic import SkipValidation

import chromadb
from chromadb import QueryResult
from chromadb.config import Settings
from chromadb.api.models.Collection import Collection

from agentuniverse.agent.action.knowledge.embedding.embedding_manager import EmbeddingManager
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.store.store import Store
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class ChromaStore(Store):
    """Object encapsulating the ChromaDB store that has vector search enabled.

    The ChromaStore object provides insert and retrieval capabilities.

    Attributes:
        collection_name (str): The name of the chroma collection to use.
        collection (Collection): A chroma collection object.
        persist_path (Optional[str]): Path to save the chroma database.
    """

    collection_name: Optional[str] = 'chroma_db'
    collection: SkipValidation[Collection] = None
    persist_path: Optional[str] = None
    embedding_model: Optional[str] = None
    similarity_top_k: Optional[int] = 10

    def _new_client(self) -> Any:
        """Initialize the chroma client."""
        if self.persist_path.startswith('http') or \
                self.persist_path.startswith('https'):
            # Remote database URL
            parsed_url = urlparse(self.persist_path)
            settings = Settings(
                chroma_api_impl="chromadb.api.fastapi.FastAPI",
                chroma_server_host=parsed_url.hostname,
                chroma_server_http_port=str(parsed_url.port)
            )
        else:
            settings = Settings(
                is_persistent=True,
                persist_directory=self.persist_path
            )

        client = chromadb.Client(settings)
        if self.collection is None:
            # default to create a new collection or get an existed collection.
            self.collection = client.get_or_create_collection(
                name=self.collection_name
            )
        return client

    def query(self, query: Query, **kwargs) -> List[Document]:
        """Query the chroma collection with the given query and return the top k results.

        Args:
            query (Query): The query object.
            **kwargs: Arbitrary keyword arguments.

        Note:
            If there is no embedding in the specific query, but the embedding model is configured in the store,
            the embedding data of the query is automatically obtained by the embedding model.

        Returns:
            List[Document]: List of documents retrieved by the query.
        """

        embedding = query.embeddings
        if self.embedding_model is not None and len(embedding) == 0:
            embedding = EmbeddingManager().get_instance_obj(
                self.embedding_model
            ).get_embeddings([query.query_str], text_type="query")[0]
        if len(embedding) > 0:
            query_result = self.collection.query(
                n_results=query.similarity_top_k if query.similarity_top_k else self.similarity_top_k,
                query_embeddings=embedding
            )
        else:
            query_result = self.collection.query(
                n_results=query.similarity_top_k if query.similarity_top_k else self.similarity_top_k,
                query_texts=[query.query_str]
            )
        # convert to the agentUniverse(aU) document format
        return self.to_documents(query_result)

    def insert_document(self, documents: List[Document], **kwargs: Any):
        """Insert documents to the chroma collection.

        Args:
            documents (List[Document]): The documents to be inserted.
            **kwargs: Arbitrary keyword arguments.

        Note:
            If there is no embedding in the specific document, but the embedding model is configured in the store,
            the embedding data of the document is automatically obtained by the embedding model.
        """

        for document in documents:
            embedding = document.embedding
            if self.embedding_model is not None and len(embedding) == 0:
                embedding = EmbeddingManager().get_instance_obj(
                    self.embedding_model
                ).get_embeddings([document.text])[0]
            self.collection.add(
                documents=[document.text],
                metadatas=[document.metadata],
                embeddings=[embedding] if len(embedding) > 0 else None,
                ids=[document.id]
            )

    def upsert_document(self, documents: List[Document], **kwargs):
        """Upsert document into the store."""
        for document in documents:
            embedding = document.embedding
            if self.embedding_model is not None and len(embedding) == 0:
                embedding = EmbeddingManager().get_instance_obj(
                    self.embedding_model
                ).get_embeddings([document.text])[0]
            self.collection.upsert(
                documents=[document.text],
                metadatas=[document.metadata],
                embeddings=[embedding] if embedding is not None else None,
                ids=[document.id]
            )

    def update_document(self, documents: List[Document], **kwargs):
        """Update document into the store."""
        for document in documents:
            embedding = document.embedding
            if self.embedding_model is not None and len(embedding) == 0:
                embedding = EmbeddingManager().get_instance_obj(
                    self.embedding_model
                ).get_embeddings([document.text])[0]
            self.collection.update(
                documents=[document.text],
                metadatas=[document.metadata],
                embeddings=[embedding] if embedding is not None else None,
                ids=[document.id]
            )

    @staticmethod
    def to_documents(query_result: QueryResult) -> List[Document]:
        """Convert the query results of ChromaDB to the agentUniverse(aU) document format."""

        if query_result is None:
            return []
        documents = []
        for i in range(len(query_result['ids'][0])):
            documents.append(Document(id=query_result['ids'][0][i],
                                      text=query_result['documents'][0][i],
                                      embedding=query_result['embeddings'][0][
                                          i]
                                      if query_result[
                                             'embeddings'] is not None else [],
                                      metadata=query_result['metadatas'][0][i]
                                      if query_result[
                                             'metadatas'] is not None else None))
        return documents

    def _initialize_by_component_configer(self,
                                          chroma_store_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(chroma_store_configer)
        if hasattr(chroma_store_configer, "collection_name"):
            self.collection_name = chroma_store_configer.collection_name
        if hasattr(chroma_store_configer, "persist_path"):
            self.persist_path = chroma_store_configer.persist_path
        if hasattr(chroma_store_configer, "embedding_model"):
            self.embedding_model = chroma_store_configer.embedding_model
        if hasattr(chroma_store_configer, "similarity_top_k"):
            self.similarity_top_k = chroma_store_configer.similarity_top_k
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 17:49
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: claude_llm.py

import asyncio
from typing import Optional, Any, Union, Iterator, AsyncIterator

import anthropic
import httpx
from langchain_core.language_models import BaseLanguageModel
from pydantic import Field

from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.claude_langchain_instance import ClaudeLangChainInstance
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_output import LLMOutput

ClaudeMAXCONTETNLENGTH = {
    "claude-3-opus-20240229": 200000,
    "claude-3-sonnet-20240229": 200000,
    "claude-3-haiku-20240307": 200000,
    "claude-2.1": 200000,
    "claude-2.0": 100000,
    "claude-instant-1.2": 100000
}


class ClaudeLLM(LLM):
    """
    This class implements an interface for interacting with the Anthropic Claude model.

    Attribute:
        anthropic_api_key (Optional[str]): The API key for the Anthropic API. Defaults to the value of the environment variable ANTHROPIC_API_KEY.
        anthropic_api_url (Optional[str]): The URL for the Anthropic API. Defaults to the value of the environment variable ANTHROPIC_API_URL.
        anthropic_proxy (Optional[str]): The proxy to use for the Anthropic API. Defaults to None.
        connection_pool_limits (Optional[int]): The maximum number of connections to keep in a pool. Defaults to None.
    """
    anthropic_api_key: Optional[str] = Field(default_factory=lambda: get_from_env('ANTHROPIC_API_KEY'))
    anthropic_api_url: Optional[str] = Field(default_factory=lambda: get_from_env('ANTHROPIC_API_URL'))
    anthropic_proxy: Optional[str] = None
    connection_pool_limits: Optional[int] = None

    def _new_client(self):
        client = anthropic.Anthropic(
            api_key=self.anthropic_api_key,
            base_url=self.anthropic_api_url,
            timeout=self.request_timeout if self.request_timeout else 60,
            max_retries=self.max_retries if self.max_retries else 2,
            http_client=httpx.Client(proxy=self.anthropic_proxy) if self.anthropic_proxy else None,
            connection_pool_limits=self.connection_pool_limits
        )
        return client

    def _new_async_client(self):
        client = anthropic.AsyncAnthropic(
            api_key=self.anthropic_api_key,
            base_url=self.anthropic_api_url,
            timeout=self.request_timeout if self.request_timeout else 60,
            max_retries=self.max_retries if self.max_retries else 2,
            http_client=httpx.AsyncClient(proxy=self.anthropic_proxy) if self.anthropic_proxy else None,
            connection_pool_limits=self.connection_pool_limits
        )
        return client

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """Run the Claude LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        self.client = self._new_client()
        chat_completion = self.client.messages.create(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            self.close()
            return self.parse_result(chat_completion)
        return self.generate_stream_result(chat_completion)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        self.client = self._new_async_client()
        chat_completion = await self.client.messages.create(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            await self.aclose()
            return self.parse_result(chat_completion)
        return self.agenerate_stream_result(chat_completion)

    @staticmethod
    def parse_result(data):
        text = data.content[0].text
        if not text:
            return
        return LLMOutput(text=text, raw=data)

    def generate_stream_result(self, chat_completion: Iterator):
        for chunk in chat_completion:
            if chunk.type != 'content_block_delta':
                continue
            yield LLMOutput(text=chunk.delta.text, raw=chunk.model_dump())
        self.close()

    async def agenerate_stream_result(self, chat_completion: AsyncIterator):
        async for chunk in chat_completion:
            print(chunk)
            if chunk.type != 'content_block_delta':
                continue
            yield LLMOutput(text=chunk.delta.text, raw=chunk.model_dump())
        await self.aclose()

    def as_langchain(self) -> BaseLanguageModel:
        """
        Convert this instance into a LangChain compatible object
        """
        return ClaudeLangChainInstance(self)

    def get_num_tokens(self, text: str) -> int:
        encode = self._new_client().count_tokens(text)
        return encode

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return ClaudeMAXCONTETNLENGTH[self.model_name]

    def close(self):
        """Close the client."""
        if hasattr(self, 'client') and self.client:
            self.client.close()

    async def aclose(self):
        """Async close the client."""
        if hasattr(self, 'async_client') and self.async_client:
            await self.async_client.aclose()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/29 16:58
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: peer_work_pattern.py
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.agent.template.executing_agent_template import ExecutingAgentTemplate
from agentuniverse.agent.template.expressing_agent_template import ExpressingAgentTemplate
from agentuniverse.agent.template.planning_agent_template import PlanningAgentTemplate
from agentuniverse.agent.template.reviewing_agent_template import ReviewingAgentTemplate
from agentuniverse.agent.work_pattern.work_pattern import WorkPattern


class PeerWorkPattern(WorkPattern):
    planning: PlanningAgentTemplate = None
    executing: ExecutingAgentTemplate = None
    expressing: ExpressingAgentTemplate = None
    reviewing: ReviewingAgentTemplate = None

    def invoke(self, input_object: InputObject, work_pattern_input: dict, **kwargs) -> dict:
        self._validate_work_pattern_members()

        peer_results = list()
        planning_result = dict()
        executing_result = dict()
        expressing_result = dict()
        reviewing_result = dict()

        retry_count = work_pattern_input.get('retry_count')
        jump_step = work_pattern_input.get('jump_step')
        eval_threshold = work_pattern_input.get('eval_threshold')

        for _ in range(retry_count):
            peer_round_results = {}

            if not planning_result or jump_step == "planning":
                planning_result = self._invoke_planning(input_object, work_pattern_input, peer_round_results)

            if not executing_result or jump_step in ["planning", "executing"]:
                executing_result = self._invoke_executing(input_object, peer_round_results)

            if not expressing_result or jump_step in ["planning", "executing", "expressing"]:
                expressing_result = self._invoke_expressing(input_object, peer_round_results)

            if not reviewing_result or jump_step in ["planning", "executing", "expressing", "reviewing"]:
                reviewing_result = self._invoke_reviewing(input_object, peer_round_results)

            peer_results.append(peer_round_results)

            if not reviewing_result or (
                    reviewing_result.get('score') and reviewing_result.get('score') >= eval_threshold):
                break

        return {'result': peer_results}

    async def async_invoke(self, input_object: InputObject, work_pattern_input: dict, **kwargs) -> dict:
        self._validate_work_pattern_members()

        peer_results = list()
        planning_result = dict()
        executing_result = dict()
        expressing_result = dict()
        reviewing_result = dict()

        retry_count = work_pattern_input.get('retry_count')
        jump_step = work_pattern_input.get('jump_step')
        eval_threshold = work_pattern_input.get('eval_threshold')

        for _ in range(retry_count):
            peer_round_results = {}
            if not planning_result or jump_step == "planning":
                planning_result = await self._async_invoke_planning(input_object, work_pattern_input,
                                                                    peer_round_results)

            if not executing_result or jump_step in ["planning", "executing"]:
                executing_result = await self._async_invoke_executing(input_object, peer_round_results)

            if not expressing_result or jump_step in ["planning", "executing", "expressing"]:
                expressing_result = await self._async_invoke_expressing(input_object, peer_round_results)

            if not reviewing_result or jump_step in ["planning", "executing", "expressing", "reviewing"]:
                reviewing_result = await self._async_invoke_reviewing(input_object, peer_round_results)

            peer_results.append(peer_round_results)

            if not reviewing_result or (
                    reviewing_result.get('score') and reviewing_result.get('score') >= eval_threshold):
                break
        return {'result': peer_results}

    def _invoke_planning(self, input_object: InputObject, agent_input: dict, peer_round_results: dict) -> dict:
        if not self.planning:
            planning_result = OutputObject({"framework": [agent_input.get('input')]})
        else:
            planning_result = self.planning.run(**input_object.to_dict())
            peer_round_results['planning_result'] = planning_result.to_dict()
        input_object.add_data('planning_result', planning_result)
        return planning_result.to_dict()

    async def _async_invoke_planning(self, input_object: InputObject, agent_input: dict,
                                     peer_round_results: dict) -> dict:
        if not self.planning:
            planning_result = OutputObject({"framework": [agent_input.get('input')]})
        else:
            planning_result = await self.planning.async_run(**input_object.to_dict())
            peer_round_results['planning_result'] = planning_result.to_dict()
        input_object.add_data('planning_result', planning_result)
        return planning_result.to_dict()

    def _invoke_executing(self, input_object: InputObject, peer_round_results: dict) -> dict:
        if not self.executing:
            executing_result = OutputObject({})
        else:
            executing_result = self.executing.run(**input_object.to_dict())
            peer_round_results['executing_result'] = executing_result.to_dict()
        input_object.add_data('executing_result', executing_result)
        return executing_result.to_dict()

    async def _async_invoke_executing(self, input_object: InputObject, peer_round_results: dict) -> dict:
        if not self.executing:
            executing_result = OutputObject({})
        else:
            executing_result = await self.executing.async_run(**input_object.to_dict())
            peer_round_results['executing_result'] = executing_result.to_dict()
        input_object.add_data('executing_result', executing_result)
        return executing_result.to_dict()

    def _invoke_expressing(self, input_object: InputObject, peer_round_results: dict) -> dict:
        if not self.expressing:
            expressing_result = OutputObject({})
        else:
            expressing_result = self.expressing.run(**input_object.to_dict())
            peer_round_results['expressing_result'] = expressing_result.to_dict()
        input_object.add_data('expressing_result', expressing_result)
        return expressing_result.to_dict()

    async def _async_invoke_expressing(self, input_object: InputObject, peer_round_results: dict) -> dict:
        if not self.expressing:
            expressing_result = OutputObject({})
        else:
            expressing_result = await self.expressing.async_run(**input_object.to_dict())
            peer_round_results['expressing_result'] = expressing_result.to_dict()
        input_object.add_data('expressing_result', expressing_result)
        return expressing_result.to_dict()

    def _invoke_reviewing(self, input_object: InputObject, peer_round_results: dict) -> dict:
        if not self.reviewing:
            reviewing_result = OutputObject({})
        else:
            reviewing_result = self.reviewing.run(**input_object.to_dict())
            peer_round_results['reviewing_result'] = reviewing_result.to_dict()
        input_object.add_data('reviewing_result', reviewing_result)
        return reviewing_result.to_dict()

    async def _async_invoke_reviewing(self, input_object: InputObject, peer_round_results: dict) -> dict:
        if not self.reviewing:
            reviewing_result = OutputObject({})
        else:
            reviewing_result = await self.reviewing.async_run(**input_object.to_dict())
            peer_round_results['reviewing_result'] = reviewing_result.to_dict()
        input_object.add_data('reviewing_result', reviewing_result)
        return reviewing_result.to_dict()

    def _validate_work_pattern_members(self):
        if self.planning and not isinstance(self.planning, PlanningAgentTemplate):
            raise ValueError(f"{self.planning} is not of the expected type AgentTemplate.")
        if self.executing and not isinstance(self.executing, ExecutingAgentTemplate):
            raise ValueError(f"{self.executing} is not of the expected type AgentTemplate.")
        if self.expressing and not isinstance(self.expressing, ExpressingAgentTemplate):
            raise ValueError(f"{self.expressing} is not of the expected type AgentTemplate.")
        if self.reviewing and not isinstance(self.reviewing, ReviewingAgentTemplate):
            raise ValueError(f"{self.reviewing} is not not of the expected type ReviewingAgentTemplate.")

    def set_by_agent_model(self, **kwargs):
        peer_work_pattern_instance = self.__class__()
        peer_work_pattern_instance.name = self.name
        peer_work_pattern_instance.description = self.description
        for key in ['planning', 'executing', 'expressing', 'reviewing']:
            if key in kwargs:
                setattr(peer_work_pattern_instance, key, kwargs[key])
        return peer_work_pattern_instance



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 20:02
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: knowledge_node.py
from typing import List, Optional

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import Node, NodeData
from agentuniverse.workflow.node.node_config import KnowledgeNodeInputParams, NodeInfoParams, NodeOutputParams
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.workflow_output import WorkflowOutput


class KnowledgeNodeData(NodeData):
    inputs: Optional[KnowledgeNodeInputParams] = None


class KnowledgeNode(Node):
    """The basic class of the knowledge node."""

    _data_cls = KnowledgeNodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.KNOWLEDGE

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        inputs: KnowledgeNodeInputParams = self._data.inputs

        param_map = {
            'top_k': None,
            'id': None
        }

        for knowledge_param in inputs.knowledge_param:
            if knowledge_param.name in param_map:
                if isinstance(knowledge_param.value, str):
                    param_map[knowledge_param.name] = knowledge_param.value
                else:
                    param_map[knowledge_param.name] = knowledge_param.value.get('content', None)

        knowledge_id_list = param_map.get('id')
        query_top_k = param_map.get('top_k')

        knowledge_list = []
        for knowledge_id in knowledge_id_list:
            knowledge: Knowledge = KnowledgeManager().get_instance_obj(knowledge_id)
            if knowledge is None:
                raise ValueError("No knowledge with id {} was found.".format(knowledge_id))
            knowledge_list.append(knowledge)

        knowledge_input_params = self._resolve_input_params(inputs.input_param, workflow_output)
        if query_top_k:
            knowledge_input_params['similarity_top_k'] = query_top_k
        if 'query' in knowledge_input_params:
            knowledge_input_params['query_str'] = knowledge_input_params.pop('query')

        knowledge_res = ""
        for knowledge in knowledge_list:
            document_texts = [document.text for document in knowledge.query_knowledge(**knowledge_input_params)]
            knowledge_res += '\n'.join(document_texts)
        output_params: List[NodeOutputParams] = self._data.outputs
        output_params[0].value = knowledge_res

        workflow_output.workflow_parameters[self.id] = output_params
        return NodeOutput(node_id=self.id, status=NodeStatusEnum.SUCCEEDED, result=output_params)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/23 11:28
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: embedding_manager.py

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase
from agentuniverse.agent.action.knowledge.embedding.embedding import Embedding


@singleton
class EmbeddingManager(ComponentManagerBase[Embedding]):
    """A singleton manager class of the embedding."""

    def __init__(self):
        super().__init__(ComponentEnum.EMBEDDING)




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/27 23:16
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool_util.py
import os
from typing import Dict

from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger
from agentuniverse.base.config.configer import Configer
from agentuniverse_product.service.model.tool_dto import ToolDTO


def validate_create_api_tool_parameters(tool_dto: ToolDTO) -> None:
    """Validate the parameters for creating an api tool instance.

    Args :
        tool_dto (ToolDTO): The tool DTO object containing the tool parameters.
    """
    if tool_dto.id is None:
        raise ValueError("Tool id cannot be None.")
    tool = ToolManager().get_instance_obj(tool_dto.id)
    if tool:
        raise ValueError(f"Tool instance corresponding to the tool id already exists. {tool_dto.id}")
    if tool_dto.openapi_schema is None:
        raise ValueError("The openapi_schema in tool cannot be None.")


def assemble_tool_product_config_data(tool_dto: ToolDTO) -> Dict:
    """Assemble the tool product configuration data.

    Args :
        tool_dto (ToolDTO): The tool DTO object containing the tool parameters.
    """
    return {
        'id': tool_dto.id,
        'nickname': tool_dto.nickname,
        'avatar': tool_dto.avatar,
        'type': 'TOOL',
        'metadata': {
            'class': 'Product',
            'module': 'agentuniverse_product.base.product',
            'type': 'PRODUCT'
        },
    }


def assemble_api_tool_config_data(tool_dto: ToolDTO) -> Dict:
    """Assemble the api tool configuration data.

    Args :
        tool_dto (ToolDTO): The tool DTO object containing the tool parameters.

    Returns:
        Dict: The assembled api tool configuration data.
    """
    tool_config_data = {
        'name': tool_dto.id,
        'description': tool_dto.description,
        'tool_type': 'api',
        'input_keys': tool_dto.parameters,
        'openapi_spec': tool_dto.openapi_schema
    }
    metadata_class = 'APITool'
    metadata_api_tool_path = 'api_tool'

    tool_config_data['metadata'] = {
        'class': metadata_class,
        'module': f'agentuniverse.agent.action.tool.{metadata_api_tool_path}',
        'type': 'TOOL'
    }
    return tool_config_data


def register_tool(file_path: str):
    """Register a tool instance to the tool manager.

    Args :
        file_path (str): The path to the tool configuration file.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    tool_configer: ToolConfiger = ToolConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(tool_configer)
    component_instance: Tool = component_clz().initialize_by_component_configer(tool_configer)
    component_instance.component_config_path = component_configer.configer.path
    ToolManager().register(component_instance.get_instance_code(), component_instance)


def parse_tool_input(openapi: dict) -> list[str]:
    """Parse the tool input from the openapi schema.

    Args:
        openapi (dict): The openapi schema dictionary.

    Returns:
        list[str]: The list of tool input.
    """
    # convert parameters
    parameters = []
    if 'parameters' in openapi['operation']:
        for parameter in openapi['operation'].get('parameters'):
            if parameter.get('required'):
                parameters.append(parameter['name'])
    return parameters



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/14 10:42
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: system_util.py

from pathlib import Path
import importlib

PROJECT_ROOT_PATH = None


def get_project_root_path() -> Path:
    """Get the project root path."""
    global PROJECT_ROOT_PATH
    if PROJECT_ROOT_PATH:
        return PROJECT_ROOT_PATH
    current_work_directory = Path.cwd()
    root_path = current_work_directory.parents[1]
    PROJECT_ROOT_PATH = root_path
    return root_path


def parse_dynamic_str(param: str):
    """
    Translate a string, firstly attempting to parse it as a full path to a
    parameterless function. If it can be correctly imported, return the result
    of the function execution, otherwise return the original string.
    This is useful in scenarios where it's inconvenient to write the real
    value directly, such as with secret keys.
    """
    try:
        parts = param.rsplit('.', 1)
        if len(parts) == 2:
            module_path, func_name = parts
            module = importlib.import_module(module_path)
            func = getattr(module, func_name)
            if callable(func):
                return func()
            else:
                return param
        else:
            return param
    except (ImportError, AttributeError, Exception):
        return param


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 14:00
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: component_configer_util.py
import importlib
from typing import Type, Callable

from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.memory.memory_compressor.memory_compressor_manager import MemoryCompressorManager
from agentuniverse.agent.memory.memory_manager import MemoryManager
from agentuniverse.agent.memory.memory_storage.memory_storage_manager import MemoryStorageManager
from agentuniverse.agent.plan.planner.planner_manager import PlannerManager
from agentuniverse.agent.work_pattern.work_pattern_manager import WorkPatternManager
from agentuniverse.agent_serve.service_manager import ServiceManager
from agentuniverse.agent_serve.service_configer import ServiceConfiger
from agentuniverse.base.config.component_configer.configers.work_pattern_configer import WorkPatternConfiger
from agentuniverse.base.config.component_configer.configers.workflow_configer import WorkflowConfiger
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.config.component_configer.configers.knowledge_configer import KnowledgeConfiger
from agentuniverse.base.config.component_configer.configers.memory_configer import MemoryConfiger
from agentuniverse.base.config.component_configer.configers.planner_configer import PlannerConfiger
from agentuniverse.base.config.component_configer.configers.prompt_configer import PromptConfiger
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger
from agentuniverse.base.config.component_configer.configers.sqldb_wrapper_config import SQLDBWrapperConfiger
from agentuniverse.base.config.config_type_enum import ConfigTypeEnum
from agentuniverse.base.config.component_configer.configers.llm_configer import LLMConfiger
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.workflow.workflow_manager import WorkflowManager

from agentuniverse.agent.action.knowledge.embedding.embedding_manager import EmbeddingManager
from agentuniverse.agent.action.knowledge.doc_processor.doc_processor_manager import DocProcessorManager
from agentuniverse.agent.action.knowledge.reader.reader_manager import ReaderManager
from agentuniverse.agent.action.knowledge.query_paraphraser.query_paraphraser_manager import QueryParaphraserManager
from agentuniverse.agent.action.knowledge.store.store_manager import StoreManager
from agentuniverse.agent.action.knowledge.rag_router.rag_router_manager import RagRouterManager


class ComponentConfigerUtil(object):
    """The ComponentConfigerUtil class, which is used to load and manage the component configuration."""

    __COMPONENT_CONFIGER_CLZ_MAP = {
        ComponentEnum.AGENT: AgentConfiger,
        ComponentEnum.KNOWLEDGE: KnowledgeConfiger,
        ComponentEnum.LLM: LLMConfiger,
        ComponentEnum.PLANNER: PlannerConfiger,
        ComponentEnum.TOOL: ToolConfiger,
        ComponentEnum.MEMORY: MemoryConfiger,
        ComponentEnum.SERVICE: ServiceConfiger,
        ComponentEnum.PROMPT: PromptConfiger,
        ComponentEnum.SQLDB_WRAPPER: SQLDBWrapperConfiger,
        ComponentEnum.WORKFLOW: WorkflowConfiger,
        ComponentEnum.EMBEDDING: ComponentConfiger,
        ComponentEnum.DOC_PROCESSOR: ComponentConfiger,
        ComponentEnum.READER: ComponentConfiger,
        ComponentEnum.STORE: ComponentConfiger,
        ComponentEnum.RAG_ROUTER: ComponentConfiger,
        ComponentEnum.QUERY_PARAPHRASER: ComponentConfiger,
        ComponentEnum.MEMORY_COMPRESSOR: ComponentConfiger,
        ComponentEnum.MEMORY_STORAGE: ComponentConfiger,
        ComponentEnum.WORK_PATTERN: WorkPatternConfiger,
        ComponentEnum.DEFAULT: ComponentConfiger
    }

    __COMPONENT_MANAGER_CLZ_MAP = {
        ComponentEnum.AGENT: AgentManager,
        ComponentEnum.KNOWLEDGE: KnowledgeManager,
        ComponentEnum.LLM: LLMManager,
        ComponentEnum.PLANNER: PlannerManager,
        ComponentEnum.TOOL: ToolManager,
        ComponentEnum.MEMORY: MemoryManager,
        ComponentEnum.SERVICE: ServiceManager,
        ComponentEnum.SQLDB_WRAPPER: SQLDBWrapperManager,
        ComponentEnum.PROMPT: PromptManager,
        ComponentEnum.WORKFLOW: WorkflowManager,
        ComponentEnum.EMBEDDING: EmbeddingManager,
        ComponentEnum.DOC_PROCESSOR: DocProcessorManager,
        ComponentEnum.READER: ReaderManager,
        ComponentEnum.STORE: StoreManager,
        ComponentEnum.RAG_ROUTER: RagRouterManager,
        ComponentEnum.QUERY_PARAPHRASER: QueryParaphraserManager,
        ComponentEnum.MEMORY_COMPRESSOR: MemoryCompressorManager,
        ComponentEnum.MEMORY_STORAGE: MemoryStorageManager,
        ComponentEnum.WORK_PATTERN: WorkPatternManager,
    }

    @classmethod
    def get_component_config_clz_by_type(cls, component_type_enum: ComponentEnum) -> \
            Type[ComponentConfiger | LLMConfiger]:
        """Get the ComponentConfiger object by the component type.
        Args:
            component_type_enum(ConfigTypeEnum): the component type
        Returns:
            ComponentConfiger: the sub object of ComponentConfiger
        """
        component_config_clz = cls.__COMPONENT_CONFIGER_CLZ_MAP.get(component_type_enum)
        if component_config_clz is None:
            raise Exception(f"Failed to get the ComponentConfiger class by the component type: {component_type_enum}")
        return component_config_clz

    @classmethod
    def get_component_object_clz_by_component_configer(cls, component_configer: ComponentConfiger) -> Callable:
        """Get the component object by the ComponentConfiger object.
        Args:
            component_configer(ComponentConfiger): the ComponentConfiger object
        Returns:
            object: the component object
        """
        module = importlib.import_module(component_configer.metadata_module)
        clz = getattr(module, component_configer.metadata_class)
        return clz

    @classmethod
    def get_component_manager_clz_by_type(cls, component_type_enum: ComponentEnum) -> Callable:
        """Get the ComponentManager object by the component type.
        Args:
            component_type_enum(ConfigTypeEnum): the component type
        Returns:
            object: the ComponentManager object
        """
        return cls.__COMPONENT_MANAGER_CLZ_MAP.get(component_type_enum)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/26 17:58
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory_configer.py
from typing import Optional, List

from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer


class MemoryConfiger(ComponentConfiger):
    """The MemoryConfiger class, which is used to load and manage the Memory configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the MemoryConfiger."""
        super().__init__(configer)
        self.__name: Optional[str] = None
        self.__description: Optional[str] = None
        self.__type: Optional[str] = None
        self.__memory_key: Optional[str] = None
        self.__max_tokens: Optional[int] = None
        self.__memory_compressor: Optional[str] = None
        self.__memory_storages: Optional[List[str]] = None
        self.__memory_retrieval_storage: Optional[str] = None

    @property
    def name(self) -> Optional[str]:
        """Return the name of the Memory."""
        return self.__name

    @property
    def description(self) -> Optional[str]:
        """Return the description of the Memory."""
        return self.__description

    @property
    def type(self) -> Optional[str]:
        """Return the type of the Memory."""
        return self.__type

    @property
    def memory_key(self) -> Optional[str]:
        """Return the key of the Memory."""
        return self.__memory_key

    @property
    def max_tokens(self) -> Optional[int]:
        """Return memory tokens of the Memory."""
        return self.__max_tokens

    @property
    def memory_compressor(self) -> Optional[str]:
        """Return the compressor of the Memory."""
        return self.__memory_compressor

    @property
    def memory_storages(self) -> Optional[List[str]]:
        """Return the storages of the Memory."""
        return self.__memory_storages

    @property
    def memory_retrieval_storage(self) -> Optional[str]:
        """Return the retrieval storage of the Memory."""
        return self.__memory_retrieval_storage

    def load(self) -> 'MemoryConfiger':
        """Load the configuration by the Configer object.
        Returns:
            MemoryConfiger: the MemoryConfiger object
        """
        return self.load_by_configer(self.__configer)

    def load_by_configer(self, configer: Configer) -> 'MemoryConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            MemoryConfiger: the MemoryConfiger object
        """
        super().load_by_configer(configer)

        try:
            self.__name = configer.value.get('name')
            self.__description = configer.value.get('description')
            self.__type = configer.value.get('type')
            self.__memory_key = configer.value.get('memory_key')
            self.__max_tokens = configer.value.get('max_tokens')
            self.__memory_compressor = configer.value.get('memory_compressor')
            self.__memory_storages = configer.value.get('memory_storages')
            self.__memory_retrieval_storage = configer.value.get('memory_retrieval_storage')
        except Exception as e:
            raise Exception(f"Failed to parse the Memory configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/11 16:02
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: framework_context_manager.py
from contextvars import ContextVar, Token
import threading
from typing import Dict, Any

from agentuniverse.base.annotation.singleton import singleton


@singleton
class FrameworkContextManager:
    """A manager class to set, get and clear thread level context."""

    def __init__(self):
        """Init an empty context variable dict and a thread lock used when
        add new key to this dict."""
        self.__context_dict: Dict[str, ContextVar] = {}
        self.__dict_edit_lock = threading.Lock()

    def is_context_exist(self, var_name: str) -> bool:
        """Judge whether context variable exist in current context.

        Args:
            var_name (`str`):
                Name of the context variable.

        Returns:
            A boolean value indicating whether the variable exists.
        """
        return var_name in self.__context_dict

    def set_context(self, var_name: str, var_value: Any) -> Token:
        """Set a context variable value.

        Args:
            var_name (`str`):
                Name of the context variable.
            var_value (`Any`):
                Value of the context variable.
        """
        if var_name not in self.__context_dict:
            with self.__dict_edit_lock:
                if var_name not in self.__context_dict:
                    self.__context_dict[var_name] = ContextVar(var_name)
        return self.__context_dict[var_name].set(var_value)

    def get_context(self,
                    var_name: str,
                    default_value: Any = None):
        """Get a context variable value.

        Args:
            var_name (`str`):
                Name of the context variable.
            default_value (`Any`, defaults to `None`):
                Value to be returned if target context variable doesn't exist.
        """
        if var_name not in self.__context_dict:
            return default_value
        return self.__context_dict[var_name].get(default_value)

    def del_context(self, var_name: str):
        """Set a context variable to None.

        Args:
            var_name (`str`): Name of the context variable.
        """
        if self.is_context_exist(var_name):
            self.__context_dict[var_name].set(None)

    def reset_context(self, var_name: str, token: Token):
        """Reset a context variable using given token.

        Args:
            var_name (`str`):
                Name of the context variable.
            token (`Token`):
                Token used to reset a context variable.
        """
        self.__context_dict[var_name].reset(token)

    def get_all_contexts(self) -> Dict[str, Any]:
        """Get all context variables and their values."""
        context_values = {}
        for var_name in self.__context_dict.keys():
            context_values[var_name] = self.get_context(var_name)
        return context_values

    def set_all_contexts(self, context_values: Dict[str, Any]):
        """Set all context variables using the provided dictionary."""
        for var_name, var_value in context_values.items():
            self.set_context(var_name, var_value)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 16:24
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_util.py
import os

from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.component_configer.configers.workflow_configer import WorkflowConfiger
from agentuniverse.base.config.configer import Configer
from agentuniverse.workflow.workflow import Workflow
from agentuniverse.workflow.workflow_manager import WorkflowManager


def register_workflow(file_path: str):
    """Register a workflow instance to the workflow manager.

    Args:
        file_path (str): The path to the specific workflow configuration file.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    workflow_configer: WorkflowConfiger = WorkflowConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(workflow_configer)
    workflow: Workflow = component_clz().initialize_by_component_configer(workflow_configer)
    workflow.component_config_path = component_configer.configer.path
    WorkflowManager().register(workflow.get_instance_code(), workflow)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/25 11:18
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: request_task.py
import asyncio
import enum
from enum import Enum
import json
import queue
import time
import uuid
from datetime import datetime, timedelta
from threading import Thread
from typing import Optional, AsyncIterator

from .dal.request_library import RequestLibrary
from .dal.entity.request_do import RequestDO
from .thread_with_result import ThreadWithReturnValue
from agentuniverse.base.util.logging.logging_util import LOGGER
from ...agent.output_object import OutputObject

EOF_SIGNAL = '{"type": "EOF"}'


@enum.unique
class TaskStateEnum(Enum):
    """All possible state of a web request task."""
    INIT = "init"
    RUNNING = "running"
    FINISHED = "finished"
    FAIL = "fail"
    CANCELED = "canceled"


# All valid transitions of request task.
VALID_TRANSITIONS = {
    (TaskStateEnum.INIT, TaskStateEnum.RUNNING),
    (TaskStateEnum.INIT, TaskStateEnum.FAIL),
    (TaskStateEnum.INIT, TaskStateEnum.CANCELED),
    (TaskStateEnum.RUNNING, TaskStateEnum.FINISHED),
    (TaskStateEnum.RUNNING, TaskStateEnum.FAIL),
    (TaskStateEnum.RUNNING, TaskStateEnum.CANCELED),
}


class RequestTask:
    """The class that manages the agent service request, including state
    transitions and different types of execution flows."""

    def __init__(self, func, saved=True, **kwargs):
        """Init a RequestTask."""
        self.func: callable = func
        self.kwargs = kwargs
        self.request_id = uuid.uuid4().hex
        self.queue = queue.Queue(maxsize=100)
        self.thread: Optional[ThreadWithReturnValue] = None
        self.state = TaskStateEnum.INIT.value
        # Whether save to Database.
        self.saved = saved
        self.__request_do__ = self.add_request_do()
        self.async_queue = asyncio.Queue(maxsize=200)
        self.async_task = None

    def receive_steps(self):
        """Yield the stream data by getting data from the queue."""
        while True:
            output: str = self.queue.get()
            if output is None:
                break
            if output == EOF_SIGNAL:
                break
            yield "data:" + json.dumps({"process": output},
                                       ensure_ascii=False) + "\n\n"
        if self.canceled():
            return
        try:
            result = self.thread.result()
            if isinstance(result, OutputObject):
                result = result.to_dict()
            yield "data:" + json.dumps({"result": result},
                                       ensure_ascii=False) + "\n\n "
        except Exception as e:
            LOGGER.error("request task execute Fail: " + str(e))
            yield "data:" + json.dumps({"error": {"error_msg": str(e)}}) + "\n\n "

    async def async_receive_steps(self) -> AsyncIterator[str]:
        while True:
            try:
                output: str = await asyncio.wait_for(self.async_queue.get(), timeout=0.5)
            except asyncio.TimeoutError:
                await asyncio.sleep(1)
                print("Waiting for data timed out. Retrying...")
                continue
            if output is None:
                break
            if output == EOF_SIGNAL:
                break
            yield "data:" + json.dumps({"process": output},
                                       ensure_ascii=False) + "\n\n"
        if self.canceled():
            return
        try:
            result = await self.async_task
            if isinstance(result, OutputObject):
                result = result.to_dict()
            yield "data:" + json.dumps({"result": result},
                                       ensure_ascii=False) + "\n\n"
        except Exception as e:
            LOGGER.error("request task execute Fail: " + str(e))
            yield "data:" + json.dumps({"error": {"error_msg": str(e)}}) + "\n\n"

    def append_steps(self):
        """Tracing async service running state and update it to database."""
        try:
            self.next_state(TaskStateEnum.RUNNING)
            while True:
                output: str = self.queue.get()
                if output is None:
                    break
                if output == EOF_SIGNAL:
                    break
                if output != "" and output != " ":
                    self.__request_do__.steps.append(output)
                if self.saved:
                    RequestLibrary().update_request(self.__request_do__)
            if self.canceled():
                self.__request_do__.result['result'] = {
                    "result": "The task's tracking status has been canceled."}
            else:
                self.__request_do__.result['result'] = self.thread.result()
                self.next_state(TaskStateEnum.FINISHED)
            if self.saved:
                RequestLibrary().update_request(self.__request_do__)
        except Exception as e:
            LOGGER.error("request task update request state Fail: " + str(e))
            self.__request_do__.result['result'] = {"error_msg": str(e)}
            self.next_state(TaskStateEnum.FAIL)
            if self.saved:
                RequestLibrary().update_request(self.__request_do__)

    def async_run(self):
        """Run the service in async mode."""
        self.kwargs['output_stream'] = self.queue
        self.thread = ThreadWithReturnValue(target=self.func,
                                            kwargs=self.kwargs)
        self.thread.start()
        Thread(target=self.append_steps).start()
        Thread(target=self.check_state).start()

    def stream_run(self):
        """Run the service in a separate thread and yield result stream."""
        self.kwargs['output_stream'] = self.queue

        self.thread = ThreadWithReturnValue(target=self.func,
                                            kwargs=self.kwargs)
        self.thread.start()
        return self.receive_steps()

    async def async_stream_run(self) -> AsyncIterator[str]:
        self.kwargs['output_stream'] = self.async_queue
        loop = asyncio.get_event_loop()
        self.async_task = loop.create_task(self.func(**self.kwargs))
        async for item in self.async_receive_steps():
            yield item

    def run(self):
        """Run the service synchronous and return the result."""
        self.next_state(TaskStateEnum.RUNNING)
        try:
            result = self.func(**self.kwargs)
            self.next_state(TaskStateEnum.FINISHED)
            self.__request_do__.result = {"result": result}
            return result
        except Exception as e:
            self.next_state(TaskStateEnum.FAIL)
            self.__request_do__.additional_args['error_msg'] = str(e)
            raise e
        finally:
            if self.saved:
                RequestLibrary().update_request(self.__request_do__)

    def next_state(self, next_state: TaskStateEnum):
        """Update request task state if the transition is valid."""
        if ((TaskStateEnum[self.__request_do__.state.upper()], next_state)
                in VALID_TRANSITIONS):
            self.__request_do__.state = next_state.value
        else:
            raise Exception("Invalid state transition")

    def check_state(self):
        """Keep check request task thread state every minute, if the thread
        is alive, update the request modified time in database."""
        while True:
            if self.thread is not None and self.thread.is_alive():
                LOGGER.debug(
                    "request:" + str(self.request_id) + "task thread alive")
                if self.saved:
                    RequestLibrary().update_gmt_modified(self.request_id)
                time.sleep(60)
                continue
            elif self.__request_do__.state == TaskStateEnum.RUNNING.value:
                # Waiting one minute to avoid skipping the state change step.
                time.sleep(60)
                if self.__request_do__.state == TaskStateEnum.RUNNING.value:
                    LOGGER.debug("request:" + str(self.request_id) +
                                 " task thread stop but state not end")
                    self.__request_do__.state = TaskStateEnum.FAIL.value
                    if self.saved:
                        RequestLibrary().update_request(
                            self.__request_do__)
            break

    def add_request_do(self):
        query_keys = ['question', 'query_content', 'query', 'request', 'input']
        query = next((self.kwargs[key] for key in query_keys if
                      self.kwargs.get(key) is not None),
                     "No relevant query was retrieved.")

        request_do = RequestDO(
            request_id=self.request_id,
            session_id="",
            query=query,
            state=TaskStateEnum.INIT.value,
            result=dict(),
            steps=[],
            additional_args=dict(),
            gmt_create=int(time.time()),
            gmt_modified=int(time.time()),
        )
        if self.saved:
            RequestLibrary().add_request(request_do)
        return request_do

    def result(self):
        """Get the result from service running thread."""
        return self.thread.result()

    @staticmethod
    def is_validate(request_do: RequestDO):
        """If there is no update within ten minutes and the status is neither
        completed nor failed, the task is considered to have failed."""
        if (request_do.gmt_modified < datetime.now() - timedelta(minutes=10)
                and request_do.state != TaskStateEnum.FINISHED.value
                and request_do.state != TaskStateEnum.FAIL.value):
            LOGGER.error("request task is validate fail" + str(request_do))
            request_do.state = TaskStateEnum.FAIL.value
            RequestLibrary().update_request(request_do)

    def cancel(self):
        """Cancel the request task. If a response SSE stream is working, put
        the EOF into the queue."""
        self.next_state(TaskStateEnum.CANCELED)
        if self.queue is not None:
            self.queue.put_nowait(EOF_SIGNAL)

    def request_state(self):
        """Return the request task state."""
        return self.__request_do__.state

    def canceled(self):
        """Whether task is canceled state."""
        return self.__request_do__.state == TaskStateEnum.CANCELED.value

    def finished(self):
        """Set task to finished state."""
        self.__request_do__.state = TaskStateEnum.FINISHED.value

    @staticmethod
    def query_request_state(request_id: str) -> dict | None:
        """Query the request data in database by given request_id.

        Args:
            request_id(str): Unique request id.
        """
        request_do = RequestLibrary().query_request_by_request_id(
            request_id)
        if request_do is None:
            return None
        RequestTask.is_validate(request_do)
        return {
            "state": request_do.state,
            "result": request_do.result,
            "steps": request_do.steps
        }



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 19:44
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: enum.py
from enum import Enum


class NodeEnum(Enum):
    """The enumeration of the supported nodes."""

    START = 'start'
    END = 'end'
    LLM = 'llm'
    TOOL = 'tool'
    KNOWLEDGE = 'knowledge'
    AGENT = 'agent'
    CONDITION = 'ifelse'

    @staticmethod
    def to_value_list():
        """Return the value list of the enumeration."""
        return [item.value for item in NodeEnum]

    @staticmethod
    def from_value(value):
        """Return the enum member corresponding to the given value."""
        for item in NodeEnum:
            if item.value == value:
                return item
        raise ValueError(f"No enum member with value {value}")


class NodeStatusEnum(Enum):
    RUNNING = 'running'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'


class ConditionComparisonEnum(Enum):
    EQUAL = 'equal'
    NOT_EQUAL = 'not_equal'
    BLANK = 'blank'



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/1 09:43
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: message_service.py
from datetime import datetime

from agentuniverse_product.dal.message_library import MessageLibrary
from agentuniverse_product.dal.model.message_do import MessageDO


class MessageService:
    """Message Service for aU-product."""

    @staticmethod
    def add_message(session_id: str, content: str, add_time: datetime) -> int:
        """Add a message to the message db table."""
        if content is None:
            raise ValueError("message content is required parameter.")
        return MessageLibrary().add_message(
            MessageDO(session_id=session_id, content=content, gmt_created=add_time, gmt_modified=add_time))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/8 14:01
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: sqlite_store.py
import sqlite3
import json
import math
from typing import List, Optional, Set
from collections import Counter

import jieba

from agentuniverse.agent.action.knowledge.store.store import Store
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.doc_processor.doc_processor_manager import DocProcessorManager
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class SQLiteStore(Store):
    db_path: str = 'sqlite_store.db'
    conn: Optional[sqlite3.Connection] = None
    k1: float = 1.5
    b: float = 0.75
    keyword_extractor: Optional[str] = None
    similarity_top_k: int = 10

    def _new_client(self):
        self.conn = sqlite3.connect(self.db_path, check_same_thread=False)
        self._create_tables()

    def _create_tables(self):
        with self.conn:
            self.conn.execute('''
                CREATE TABLE IF NOT EXISTS documents (
                    id TEXT PRIMARY KEY,
                    text TEXT,
                    word_count INT,
                    metadata TEXT
                )
            ''')
            self.conn.execute('''
                CREATE TABLE IF NOT EXISTS inverted_index (
                    term TEXT,
                    doc_id TEXT,
                    FOREIGN KEY (doc_id) REFERENCES documents (id)
                )
            ''')

    def _initialize_by_component_configer(self,
                                          sqlite_store_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(sqlite_store_configer)
        if hasattr(sqlite_store_configer, "db_path"):
            self.db_path = sqlite_store_configer.db_path
        if hasattr(sqlite_store_configer, "k1"):
            self.k1 = sqlite_store_configer.k1
        if hasattr(sqlite_store_configer, "b"):
            self.b = sqlite_store_configer.b
        if hasattr(sqlite_store_configer, "keyword_extractor"):
            self.keyword_extractor = sqlite_store_configer.keyword_extractor
        if hasattr(sqlite_store_configer, "similarity_top_k"):
            self.similarity_top_k = sqlite_store_configer.similarity_top_k
        return self


    def _get_all_docs_count(self) -> int:
        with self.conn:
            cursor = self.conn.cursor()
            cursor.execute('SELECT COUNT(*) FROM documents')
            count = cursor.fetchone()[0]
            cursor.close()
        return count

    def _get_all_docs_words_count(self) -> int:
        with self.conn:
            cursor = self.conn.cursor()
            cursor.execute('SELECT SUM(word_count) FROM documents')
            total_word_count = cursor.fetchone()[0]
            cursor.close()
        return total_word_count if total_word_count is not None else 0

    def compute_bm25(self, query_text, doc_text, inverted_index,
                     total_doc_count, total_word_count):
        k1 = self.k1
        b = self.b
        query_words = jieba.lcut(query_text)
        doc_words = jieba.lcut(doc_text)
        doc_length = len(doc_words)
        avg_doc_length = total_word_count / total_doc_count
        bm25_score = 0

        doc_counter = Counter(doc_words)

        for term in query_words:
            if term in inverted_index:
                term_freq = doc_counter[term]
                num_docs_with_term = len(inverted_index[term])
                idf = math.log((total_doc_count - num_docs_with_term + 0.5) / (
                        num_docs_with_term + 0.5) + 1)
                bm25_term_score = idf * (term_freq * (k1 + 1)) / (
                        term_freq + k1 * (
                        1 - b + b * (doc_length / avg_doc_length)))
                bm25_score += bm25_term_score

        return bm25_score


    def _get_document_keyword(self, document: Document) -> Set[str]:
        if not self.keyword_extractor:
            raise Exception(
                "You must specify a keyword extractor in sqlite store query")
        else:
            _doc = DocProcessorManager().get_instance_obj(
                self.keyword_extractor) \
                .process_docs([document])
            return _doc[0].keywords

    def insert_document(self, documents: List[Document], **kwargs):
        with self.conn:
            for document in documents:
                metadata = json.dumps(
                    document.metadata) if document.metadata else None
                self.conn.execute(
                    'INSERT OR REPLACE INTO documents (id, text, word_count, metadata) VALUES (?, ?, ?, ?)',
                    (document.id, document.text, len(jieba.lcut(document.text)), metadata)
                )
                self._get_document_keyword(document)
                for term in set(document.keywords):
                    self.conn.execute(
                        'INSERT INTO inverted_index (term, doc_id) VALUES (?, ?)',
                        (term, document.id)
                    )

    def delete_document(self, document_id: int):
        with self.conn:
            self.conn.execute(
                'DELETE FROM documents WHERE id = ?',
                (document_id,)
            )
            self.conn.execute(
                'DELETE FROM inverted_index WHERE doc_id = ?',
                (document_id,)
            )

    def upsert_document(self, documents: List[Document], **kwargs):
        with self.conn:
            for document in documents:
                metadata = json.dumps(
                    document.metadata) if document.metadata else None
                self.conn.execute(
                    'INSERT OR REPLACE INTO documents (id, text, word_count, metadata) VALUES (?, ?, ?, ?)',
                    (
                    document.id, document.text, len(jieba.lcut(document.text)),
                    metadata)
                )
                self.conn.execute(
                    'DELETE FROM inverted_index WHERE doc_id = ?',
                    (document.id,)
                )
                self._get_document_keyword(document)
                for term in set(document.keywords):
                    self.conn.execute(
                        'INSERT INTO inverted_index (term, doc_id) VALUES (?, ?)',
                        (term, document.id)
                    )

    def query(self, query: Query, **kwargs) -> List[Document]:
        if len(query.keywords) > 0:
            query_terms = query.keywords
        else:
            query_terms = self._get_document_keyword(Document(text=query.query_str))
            query.keywords = query_terms

        # Get doc_id from inverted index.
        relevant_docs = set()
        inverted_index = {}
        with self.conn:
            for keyword in query_terms:
                cursor = self.conn.cursor()
                cursor.execute(
                    'SELECT doc_id FROM inverted_index WHERE term = ?',
                    (keyword,))
                docs_with_keyword = [row[0] for row in cursor.fetchall()]
                inverted_index[keyword] = docs_with_keyword
                relevant_docs.update(docs_with_keyword)
                cursor.close()

        # Count every document's bm25.
        doc_scores = []
        total_doc_count = self._get_all_docs_count()
        total_word_count = self._get_all_docs_words_count()
        for doc_id in relevant_docs:
            cursor = self.conn.cursor()
            cursor.execute('SELECT text FROM documents WHERE id = ?',
                           (doc_id,))
            doc_text = cursor.fetchone()[0]
            cursor.close()

            bm25_score = self.compute_bm25(query.query_str, doc_text,
                                           inverted_index, total_doc_count, total_word_count)
            doc_scores.append((doc_id, bm25_score))

        # Order the docs with bm25, and return top k.
        doc_scores.sort(key=lambda x: x[1], reverse=True)
        top_docs = doc_scores[:self.similarity_top_k]
        results = []
        for doc_id, score in top_docs:
            cursor = self.conn.cursor()
            cursor.execute('SELECT * FROM documents WHERE id = ?',
                           (doc_id,))
            doc_row = cursor.fetchone()
            cursor.close()

            document = Document(id=doc_row[0], text=doc_row[1],
                                word_count=doc_row[2],
                                metadata=json.loads(doc_row[3]))
            results.append(document)

        return results

    @staticmethod
    def to_documents(query_result) -> List[Document]:
        """Convert the query results of sqlite to the agentUniverse(aU) document format."""

        if query_result is None:
            return []
        documents = []
        for i in range(len(query_result['ids'][0])):
            documents.append(Document(id=query_result['ids'][0][i],
                                      text=query_result['documents'][0][i],
                                      embedding=[],
                                      metadata=json.loads(query_result[2]) if query_result[2] else None))
        return documents



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/12 17:57
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: planner_product.py
from typing import Optional, List

from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_configer import ProductConfiger


class PlannerProduct(Product):
    """The basic class of the planner product."""

    member_keys: Optional[List[str]] = None
    _instance: Optional[Planner] = None

    @property
    def instance(self) -> Planner:
        return self._instance

    def initialize_by_component_configer(self,
                                         product_configer: ProductConfiger) -> 'Product':
        """Initialize the Product by the ProductConfiger object.

        Args:
            product_configer(ProductConfiger): A configer contains product basic info.
        Returns:
            Product: A Product instance.
        """
        super().initialize_by_component_configer(product_configer)
        if hasattr(product_configer, 'member_keys') and product_configer.member_keys:
            self.member_keys = product_configer.member_keys
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/23 18:10
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: gunicorn_server.py

import tomli
from gunicorn.app.base import BaseApplication

from .flask_server import app
from .post_fork_queue import POST_FORK_QUEUE
from ...base.annotation.singleton import singleton


DEFAULT_GUNICORN_CONFIG = {
    'bind': '127.0.0.1:8888',
    'workers': 5,
    'backlog': 2048,
    'worker_class': 'gthread',
    'threads': 4,
    'timeout': 60,
    'keepalive': 10
}


# Execute all func in the queue after fork chile process.
def post_fork(server, worker):
    for _func, args, kwargs in POST_FORK_QUEUE:
        _func(*args, **kwargs)


@singleton
class GunicornApplication(BaseApplication):
    """Use gunicorn to wrap the flask web server."""
    def __init__(self, config_path: str = None):
        self.options = {}
        if config_path:
            self.__load_config_from_file(config_path)
        else:
            self.default_config = None
        self.application = app
        super().__init__()

    def load_config(self):
        """Check the config file first, use default config while config file
        not exist, then overwrite parts which in options."""
        if not self.default_config:
            config = DEFAULT_GUNICORN_CONFIG
        else:
            config = self.default_config
        for key, value in config.items():
            if key in self.cfg.settings and value is not None:
                self.cfg.set(key.lower(), value)

        # The priority of the passed arguments supersedes that of config file.
        for key, value in self.options.items():
            if key in self.cfg.settings and value is not None:
                self.cfg.set(key.lower(), value)

        # Set post fork.
        self.cfg.set('post_fork', post_fork)

    def update_config(self, options: dict):
        self.options = options
        self.load_config()

    def load(self):
        return self.application

    def __load_config_from_file(self, config_path: str):
        """Load gunicorn config file."""
        try:
            with open(config_path, 'rb') as f:
                config = tomli.load(f)["GUNICORN_CONFIG"]
        except (FileNotFoundError, TypeError):
            print("can't find gunicorn config file, use default config")
            return
        except (tomli.TOMLDecodeError, KeyError):
            print("gunicorn config file isn't a valid toml, "
                  "use default config.")
            return

        self.default_config = {
            key: value for key, value in config.items()
        }